from __future__ import annotations

import functools
from typing import TYPE_CHECKING, TypedDict, TypeVar

import pydantic

from compass.core.settings import Settings

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Hashable

    T = TypeVar("T", bound=object)
    C = Callable[..., T]


def default_set(_key: tuple[str, int | Hashable], value: T, /) -> T:  # default no-op
    return value


def default_get(_key: tuple[str, int | Hashable]) -> T | None:  # default no-op
    return None


class _OptClass(TypedDict):  # avoid global scope
    disable_cache: bool
    set_cache: Callable[[tuple[str, int | Hashable], T], T]
    get_cache: Callable[[tuple[str, int | Hashable]], T | None]


_Opt: _OptClass = {"disable_cache": True, "set_cache": default_set, "get_cache": default_get}


def setup_cache_hooks(
    set_cache: Callable[[tuple[str, int | Hashable], T], T],
    get_cache: Callable[[tuple[str, int | Hashable]], T | None],
    disable_cache: bool = False,
) -> None:
    """Turn on caching and set options.

    Args:
        set_cache: Function to set value to given key
        get_cache: Function to retrieve value from a given key
        disable_cache: Local disable of cache, for usage by backends

    """
    Settings.use_cache = True
    _Opt["disable_cache"] = disable_cache
    _Opt["set_cache"] = set_cache
    _Opt["get_cache"] = get_cache


def cache_result(*, key: tuple[str, int], model_type: type | None = None) -> Callable[[C[T]], C[T]]:
    def decorating_function(user_function: C[T]) -> C[T]:
        return _cache_wrapper(user_function, key, model_type)

    return decorating_function


def _cache_wrapper(user_function: C[T], key: tuple[str, int], model_type: type | None = None) -> C[T]:
    if Settings.use_cache is False or _Opt["disable_cache"] == 0:  # No caching
        return user_function

    def wrapper(*args: Hashable, **kwargs: Hashable) -> T:
        key_ = key[0], (*args, *kwargs)[key[1]]  # second tuple item is position of key_id in function args
        result = _Opt["get_cache"](key_)  # pylint: disable=assignment-from-none  # default returns None, backends can override
        if result is not None:
            if model_type is not None:
                return pydantic.parse_obj_as(model_type, result)
            return result
        result = user_function(*args, **kwargs)
        _Opt["set_cache"](key_, result)
        return result

    return functools.update_wrapper(wrapper, user_function)



"""Compass Interface - Core!

This module exposes the public api for CI core
"""
from __future__ import annotations

from compass.core import logger
from compass.core.errors import CompassAuthenticationError
from compass.core.errors import CompassError
from compass.core.errors import CompassNetworkError
from compass.core.errors import CompassPermissionError
from compass.core.errors import CompassReportError
from compass.core.errors import CompassReportPermissionError
from compass.core.hierarchy import Hierarchy
from compass.core.logon import Logon
from compass.core.people import People
from compass.core.reports import Reports
from compass.core.reports import TYPES_REPORTS
from compass.core.schemas.hierarchy import DescendantData
from compass.core.schemas.hierarchy import HierarchyLevel
from compass.core.schemas.hierarchy import HierarchyMember
from compass.core.schemas.hierarchy import HierarchySection
from compass.core.schemas.hierarchy import HierarchyUnit
from compass.core.schemas.hierarchy import HierarchyUnitMembers
from compass.core.schemas.hierarchy import TYPES_HIERARCHY_LEVELS
from compass.core.schemas.hierarchy import TYPES_UNIT_LEVELS
from compass.core.schemas.hierarchy import UnitData
from compass.core.schemas.logon import CompassProps
from compass.core.schemas.logon import TYPES_ORG_LEVELS
from compass.core.schemas.logon import TYPES_ROLE
from compass.core.schemas.member import AddressData
from compass.core.schemas.member import MemberAward
from compass.core.schemas.member import MemberDetails
from compass.core.schemas.member import MemberDisclosure
from compass.core.schemas.member import MemberMandatoryTraining
from compass.core.schemas.member import MemberPermit
from compass.core.schemas.member import MemberRoleCore
from compass.core.schemas.member import MemberRolePopup
from compass.core.schemas.member import MemberRolesCollection
from compass.core.schemas.member import MemberTrainingTab
from compass.core.schemas.member import TYPES_ROLE_STATUS
from compass.core.settings import Settings


class CompassInterface:
    def __init__(self, user_props: Logon, /):
        """This is the main (programmatic) interface to CI core."""
        self.user = user_props
        self.people = People(user_props)
        self.hierarchy = Hierarchy(user_props)
        self.reports = Reports(user_props)


def login(username: str, password: str, /, *, role: str | None = None, location: str | None = None) -> CompassInterface:
    """Log in to compass, return a compass.logon.Logon object.

    This function is provided as a convenient interface to the logon module.
    """
    return CompassInterface(Logon.from_logon((username, password), role, location))


__all__ = (
    # public sub-modules
    "logger",
    # public classes
    "CompassInterface",
    "Logon",
    "Hierarchy",
    "People",
    "Reports",
    "Settings",
    # public exceptions
    "CompassError",
    "CompassAuthenticationError",
    "CompassNetworkError",
    "CompassPermissionError",
    "CompassReportError",
    "CompassReportPermissionError",
    # public functions
    "login",
    # public types
    "TYPES_HIERARCHY_LEVELS",
    "TYPES_ORG_LEVELS",
    "TYPES_REPORTS",
    "TYPES_ROLE",
    "TYPES_ROLE_STATUS",
    "TYPES_UNIT_LEVELS",
    "AddressData",
    "CompassProps",
    "DescendantData",
    "HierarchyLevel",
    "HierarchyMember",
    "HierarchySection",
    "HierarchyUnit",
    "HierarchyUnitMembers",
    "MemberAward",
    "MemberDetails",
    "MemberDisclosure",
    "MemberMandatoryTraining",
    "MemberPermit",
    "MemberRoleCore",
    "MemberRolePopup",
    "MemberRolesCollection",
    "MemberTrainingTab",
    "UnitData",
)



"""Functions to directly interface with Compass operations to extract member data.

Compass's MemberProfile.aspx has 13 tabs:
 1. Personal Details (No Key)
 2. Your Children (Page=CHILD)
 3. Roles (Page=ROLES)
 4. Permits (Page=PERMITS)
 5. Training (Page=TRAINING)
 6. Awards (Page=AWARDS)
 7. Youth Badges/Awards (Page=BADGES)
 8. Event Invitations (Page=EVENTS)
 9. Emergency Details (Page=EMERGENCY)
 10. Communications (Page=COMMS)
 11. Visibility (Page=VISIBILITY)
 12. Disclosures (Page=DISCLOSURES)
 13. Parents/Guardians (Page=PARENT)

Of these, tabs 2, 7, 8, 13 are disabled functionality.
Tab 11 (Visibility) is only shown on the members' own profile.

For member-adjacent operations there are additional endpoints:
 - /Popups/Profile/AssignNewRole.aspx
 - /Popups/Maint/NewPermit.aspx
 - /Popups/Profile/EditProfile.aspx

Currently we only use one of these endpoints (AssignNewRole), as all
other data we need can be found from the MemberProfile tabs.

"""

from __future__ import annotations

import datetime
import re
from typing import Literal, Optional, TYPE_CHECKING, Union

from lxml import html

import compass.core as ci
from compass.core.logger import logger
from compass.core.settings import Settings
from compass.core.util import cache_hooks
from compass.core.util.context_managers import validation_errors_logging
from compass.core.util.type_coercion import maybe_int
from compass.core.util.type_coercion import parse_date

if TYPE_CHECKING:
    from collections.abc import Collection
    from collections.abc import Iterator

    from compass.core.util.client import Client

    TYPES_OTHER_SECTIONS = Literal["Disabilities", "Qualifications", "Hobbies"]
    TYPES_TRAINING_MODULE = dict[str, Union[None, int, str, datetime.date]]
    TYPES_TRAINING_PLPS = dict[int, list[TYPES_TRAINING_MODULE]]
    TYPES_TRAINING_OGL_DATES = dict[Literal["completed_date", "renewal_date"], Optional[datetime.date]]
    TYPES_TRAINING_OGL = dict[str, TYPES_TRAINING_OGL_DATES]

# _get_member_profile_tab
MEMBER_PROFILE_TAB_TYPES = Literal[
    "Personal", "Roles", "Permits", "Training", "Awards", "Emergency", "Comms", "Visibility", "Disclosures"
]  # sync with below
TABS = {"PERSONAL", "ROLES", "PERMITS", "TRAINING", "AWARDS", "EMERGENCY", "COMMS", "VISIBILITY", "DISCLOSURES"}  # sync with above

# get_roles_tab
ROLE_STATUSES = {"Cancelled", "Closed", "Full", "Pre provisional", "Provisional"}  # sync with schemas/member.py::TYPES_ROLE_STATUS
NON_VOLUNTEER_TITLES = {
    # occasional helper roles:
    "group occasional helper",
    "group occasional helper.",
    "district occasional helper",
    "county occasional helper",
    "country occasional helper",
    # pvg roles:
    "group non member pvg only"
    "group non member pvg only (prov)"
    "group non member pvg only (pre-prov)"
    "group non memberdetails pvg only"
    "district non member pvg only"
    "scotland region non-member pvg"
    "non member pvg only"
    "non member pvg only (prov)"
    "non member pvg only (pre-prov)"
    # council roles:
    "county scout council member",
    "county scout council member - nominated representative",
    "county scout council member - nominated youth representative",
    "county scout council member - nominated member (18-24)",
    "country scout council member - nominated representative",
    "scotland region scout council member",
    "scotland region scout council member - nominated representative ",
    "scotland region scout council member - nominated youth representative ",
    "tsa council member",
    "tsa council member - elected member",
    "tsa council member - ex-officio member",
    "tsa council member - nominated member",
    "tsa council member - nominated member (18-24)",
    # staff roles:
    "district staff",
    "county staff",
    # network member roles:
    "network member",  # TODO is this ever a role title?
    "scout network member",
    "district scout network",
    "district scout network member",
    "county scout network member",
    "bso scout network member",
}

# _reduce_date_list
ADJACENT = datetime.timedelta(days=-1)

# get_training_tab
mogl_modules = {
    "SA": "safety",
    "SG": "safeguarding",
    "FA": "first_aid",
    "gdpr": "gdpr",  # phony entry, used for key
}


def _get_member_profile_tab(client: Client, membership_number: int, profile_tab: MEMBER_PROFILE_TAB_TYPES) -> html.HtmlElement:
    """Returns data from a given tab in MemberProfile for a given member.

    Args:
        client: HTTP client
        membership_number: Membership Number to use
        profile_tab: Tab requested from Compass

    Returns:
        Response from the remote server as bytes serialised blob

    Raises:
        CompassNetworkError:
            For errors while executing the HTTP call
        CompassError: The given profile_tab value is illegal

    """
    tab_upper: str = profile_tab.upper()  # No longer type MEMBER_PROFILE_TAB_TYPES as upper case
    url = f"{Settings.base_url}/MemberProfile.aspx?CN={membership_number}"
    if tab_upper in TABS:
        url += f"&Page={tab_upper}&TAB"
    elif tab_upper != "PERSONAL":  # Personal tab has no key so is a special case
        raise ci.CompassError(f"Specified member profile tab {profile_tab} is invalid. Allowed values are {TABS}")

    response = client.get(url).content
    tree = html.fromstring(response)
    if tree.forms[0].action == "./ScoutsPortal.aspx?Invalid=AccessCN":
        raise ci.CompassPermissionError(f"You do not have permission to the details of {membership_number}")
    return tree


@cache_hooks.cache_result(key=("personal", 1))
def get_personal_tab(client: Client, membership_number: int, /) -> ci.MemberDetails:
    """Returns data from Personal Details tab for a given member.

    Args:
        client: HTTP client
        membership_number: Membership Number to use

    Returns:
        A MemberDetails object keys to the corresponding data from the personal
        data tab:

        MemberDetails(
            membership_number=...,
            name="...",
            known_as="...",
            forenames="...",
            surname="...",
            birth_date=datetime.date(...),
            sex="...",
            nationality="...",
            ethnicity="...",
            religion="...",
            occupation="...",
            join_date=datetime.date(...),
            postcode="...",
            main_phone="...",
            main_email="..."
            address=MemberDetailsAddress(...)
        )

        Keys will be present only if valid data could be extracted and
        parsed from Compass.

    Raises:
        CompassNetworkError:
            For errors while executing the HTTP call
        CompassPermissionError:
            Access to the member is not given by the current authentication

    """
    tree = _get_member_profile_tab(client, membership_number, "Personal")
    details: dict[str, Union[None, int, str, datetime.date, ci.AddressData, dict[str, str]]] = dict()

    # ### Extractors
    # ## Core:
    details["membership_number"] = membership_number
    names = next(tree.iter("title")).text.strip().split(" ")[3:]  # ("Scout", "-", membership_number, *names)
    details["forenames"] = names[0]
    details["surname"] = " ".join(names[1:])

    # Full path is:
    # XP:   /html/body/form/div[5]/div[1]/div[2]/div[2]/div/div
    # CSS:  html body form div div#mstr_container div#mstr_panel div#mstr_scroll div#mstr_work div#mpage1.mpage
    div_profile_tbl = tree[1][0][5][0][1][2][0][0]
    personal_details: dict[str, str] = {row[0][0].text: row[1][0].text for row in div_profile_tbl[1][2][1][1][0]}
    if len(div_profile_tbl) >= 5:
        contact_details: dict[str, str] = {row[0][0][0].text: row[2][0].text for row in div_profile_tbl[5][2] if row[0]}
    else:
        contact_details = {}

    # ## Core - Positional:
    details["name"] = personal_details["Name:"]  # Full Name
    details["known_as"] = personal_details["Known As:"]
    join_date = personal_details["Date of Joining:"]
    details["join_date"] = parse_date(join_date) if join_date != "Unknown" else None

    # ## Core - Position Varies:
    details["sex"] = personal_details["Gender:"]

    # ## Additional - Position Varies, visible for most roles:
    details["address"] = _process_address(contact_details.get("Address", ""))
    details["main_phone"] = contact_details.get("Phone")
    details["main_email"] = contact_details.get("Email")

    # ## Additional - Position Varies, visible for admin roles (Manager, Administrator etc):
    details["birth_date"] = parse_date(personal_details.get("Date of Birth:", ""))
    details["nationality"] = personal_details.get("Nationality:", "").strip()
    details["ethnicity"] = _process_extra(personal_details.get("Ethnicity:", ""))[0]  # discard ethnicity detail, could keep?
    details["religion"], details["religion_detail"] = _process_extra(personal_details.get("Religion/Faith:", ""))
    details["occupation"], details["occupation_detail"] = _process_extra(personal_details.get("Occupation:", ""))

    # ## Other Sections
    if len(div_profile_tbl) >= 17:  # As far as I know, the `other' sections are all-or-nothing, so we can skip to check hobbies
        details["disabilities"] = _process_misc_sections(div_profile_tbl[9][2], "Disabilities")
        details["qualifications"] = _process_misc_sections(div_profile_tbl[13][2], "Qualifications")
        details["hobbies"] = _process_misc_sections(div_profile_tbl[17][2], "Hobbies")

    # Filter out keys with no value.
    details = {k: v for k, v in details.items() if v}

    with validation_errors_logging(membership_number):
        return ci.MemberDetails.parse_obj(details)


def _process_address(address: str) -> ci.AddressData:
    if "UK" in address:
        addr_main, addr_code = address.rsplit(". ", 1)
        postcode, country = addr_code.rsplit(" ", 1)  # Split Postcode & Country
        try:
            street, town, county = addr_main.rsplit(", ", 2)  # Split address lines
            return dict(unparsed_address=address, country=country, postcode=postcode, county=county, town=town, street=street)
        except ValueError:
            street, town = addr_main.rsplit(", ", 1)
            return dict(unparsed_address=address, country=country, postcode=postcode, county=None, town=town, street=street)
    return dict(unparsed_address=address, country=None, postcode=None, county=None, town=None, street=None)


def _process_extra(field: str) -> tuple[str, Optional[str]]:
    if " - " in field:
        field, _, extra = field.strip().partition(" - ")
        return field, extra
    return field.strip(), None


def _process_misc_sections(section_table: html.HtmlElement, section_type: TYPES_OTHER_SECTIONS) -> dict[str, str]:
    out = {}
    for row in section_table:
        field, _, optional_detail = row[0][0].text.partition(" - ")
        out[field] = optional_detail

    # Filter out no-info sections. After the filter as empty here means we found the section, but it was empty
    if len(out) == 1 and f"No {section_type} Entered" in out:
        return {}
    return out


@cache_hooks.cache_result(key=("roles", 1))
def get_roles_tab(
    client: Client,
    membership_number: int,
    /,
    only_volunteer_roles: bool = True,
) -> ci.MemberRolesCollection:
    """Returns data from Roles tab for a given member.

    Parses the data to a common format, and removes Occasional Helper, PVG,
    Network, Council and Staff roles by default.

    Args:
        client: HTTP client
        membership_number: Membership Number to use
        only_volunteer_roles: If True, drop non-volunteer roles

    Returns:
        A MemberRolesCollection object with the data from the roles tab
        (keys will always be present):

        MemberRolesCollection(
            roles={
                1234578: MemberRoleCore(
                    role_number=...,
                    membership_number=...,
                    role_title='...',
                    role_class='...',
                    role_type='...',
                    location_id=...,
                    location_name='...',
                    role_start=datetime.date(...),
                    role_end=datetime.date(...),
                    role_status='...',
                    review_date=datetime.date(...),
                    can_view_details=True|False
                ),
                ...
            },
            membership_duration=...,
            primary_role=...,
        )

    Raises:
        CompassNetworkError:
            For errors while executing the HTTP call
        CompassPermissionError:
            Access to the member is not given by the current authentication

    """
    logger.debug(f"getting roles tab for member number: {membership_number}")

    rows = _get_member_profile_tab(client, membership_number, "Roles").xpath("//tbody/tr")

    primary_role = None
    roles_dates = []
    roles_data = {}
    for row in rows:
        # Get children (cells in row)
        cells = list(row)
        # If current role allows selection of role for editing, remove tickbox
        # If any role allows for selection, an additional column will be added
        # with empty table-cells where there is no tickbox. Also remove these.
        if any(el.tag == "input" for el in cells[0]) or cells[0].getchildren() == []:
            cells.pop(0)

        role_class = cells[1].text_content().strip()
        # sanitise primary roles and note primary role status
        role_title, primary_role = _extract_primary_role(cells[0].text_content().strip(), primary_role)
        # Skip OHs etc as early as possible
        non_volunteer = "helper" in role_class.lower() or role_title.lower() in NON_VOLUNTEER_TITLES
        if only_volunteer_roles and non_volunteer:
            continue

        # process role status & review date string
        role_status, review_date = _extract_review_date(cells[5].text_content().strip())

        role_details = ci.MemberRoleCore(
            role_number=int(row.get("data-pk")),
            membership_number=membership_number,
            role_title=role_title,
            role_class=role_class,
            # role_type only visible if access to System Admin tab
            role_type=cells[0][0].get("title", None),
            # location_id only visible if role is in hierarchy AND location still exists
            location_id=cells[2][0].get("data-ng_id"),
            location_name=cells[2].text_content().strip(),
            role_start=parse_date(cells[3].text_content().strip()),  # type  ignore[arg-type]
            role_end=parse_date(cells[4].text_content().strip()),
            role_status=role_status,
            review_date=review_date,
            can_view_details=any("VIEWROLE" in el.get("class") for el in cells[6]),
        )
        # Save role number if role is primary
        if primary_role is True:
            primary_role = role_details.role_number
        # If role is a full volunteer role, potentially add to date list
        if role_status != "Cancelled" and not non_volunteer:
            # If role_end is a falsy value (None), replace with today's date
            roles_dates.append((role_details.role_start, role_details.role_end or datetime.date.today()))

        roles_data[role_details.role_number] = role_details

    with validation_errors_logging(membership_number):
        # Calculate days of membership (inclusive), normalise to years.
        return ci.MemberRolesCollection(
            roles=roles_data,
            membership_duration=_membership_duration(roles_dates),
            primary_role=primary_role,
        )


def _extract_primary_role(role_title: str, primary_role: Union[int, None]) -> tuple[str, Union[int, Literal[True], None]]:
    if primary_role is not None:
        return role_title, primary_role
    if role_title.endswith(" [Primary]"):
        return role_title.removesuffix(" [Primary]"), True
    return role_title, primary_role


def _extract_review_date(review_status: str) -> tuple[ci.TYPES_ROLE_STATUS, Optional[datetime.date]]:
    if review_status.startswith("Full Review Due ") or review_status.startswith("Full Ending "):
        role_status: Literal["Full"] = "Full"
        review_date = parse_date(review_status.removeprefix("Full Review Due ").removeprefix("Full Ending "))
        return role_status, review_date
    if review_status in ROLE_STATUSES:
        return review_status, None  # type: ignore[return-value]  # str -> Literal type narrowing doesn't seem to work?
    raise ci.CompassError(f"Invalid value for review status '{review_status}'!")


def _membership_duration(dates: Collection[tuple[datetime.date, datetime.date]]) -> float:
    """Calculate days of membership (inclusive), normalise to years."""
    if not dates:  # handle empty iterable
        return 0
    membership_duration_days = sum((end - start).days + 1 for start, end in _reduce_date_list(dates))
    return round(membership_duration_days / 365.2425, 3)  # Leap year except thrice per 400 years.


def _reduce_date_list(dates: Collection[tuple[datetime.date, datetime.date]]) -> Iterator[tuple[datetime.date, datetime.date]]:
    """Reduce list of start and end dates to disjoint ranges.

    Iterate through date pairs and get longest consecutive date ranges. Returns
    all found date pairs.

    Args:
        dates: list of start, end date pairs

    Returns:
        list of reduced start, end date pairs

    """
    if len(dates) == 0:
        return []  # bail

    dates = sorted(dates)  # sort dates oldest first
    reduced = [dates[0]]  # initialise reduced list with oldest start date
    for start, end in dates[1:]:
        range_start, range_end = reduced[-1]
        # if the start date is more than one day apart from the previous end,
        # start a new date range. This allow dates to be adjacent, for example
        # leaving a role on the 31st and starting the next on the 1st
        if (range_end - start) < ADJACENT:
            reduced.append((start, end))
        # if the potential end date is newer than the current end date, modify
        # the current date range to use the new end date
        elif end > range_end:
            reduced[-1] = range_start, end

    return reduced


@cache_hooks.cache_result(key=("permits", 1))
def get_permits_tab(client: Client, membership_number: int, /) -> list[ci.MemberPermit]:
    """Returns data from Permits tab for a given member.

    If a permit has been revoked, the expires value is None and the status is PERM_REV

    Args:
        client: HTTP client
        membership_number: Membership Number to use

    Returns:
        A list of dicts mapping keys to the corresponding data from the
        permits tab (keys will always be present).

    Raises:
        CompassNetworkError:
            For errors while executing the HTTP call

    """
    tree = _get_member_profile_tab(client, membership_number, "Permits")
    # Get rows with permit content
    rows = tree.xpath('//table[@id="tbl_p4_permits"]//tr[@class="msTR msTRPERM"]')

    permits = []
    with validation_errors_logging(membership_number):
        for row in rows:
            child_nodes = list(row)
            expires = child_nodes[5].text_content()
            permits.append(
                ci.MemberPermit(
                    membership_number=membership_number,
                    permit_type=child_nodes[1].text_content(),
                    category=child_nodes[2].text_content(),
                    type=child_nodes[3].text_content(),
                    restrictions=child_nodes[4].text_content(),
                    expires=parse_date(expires) if expires != "Revoked" else None,
                    status=child_nodes[5].get("class"),
                )
            )
    return permits


@cache_hooks.cache_result(key=("training", 1))
def get_training_tab(client: Client, membership_number: int, /) -> ci.MemberTrainingTab:
    """Returns data from Training tab for a given member.

    Args:
        client: HTTP client
        membership_number: Membership Number to use

    Returns:
        A model with the data from the training tab (keys will always be
        present):

        {'roles': {1234567: {'role_number': 1234567,
           'role_title': '...',
           'role_start': datetime.datetime(...),
           'role_status': '...',
           'location': '...',
           'ta_data': '...',
           'ta_number': '...',
           'ta_name': '...',
           'completion': '...',
           'wood_badge_number': '...'},
          ...},
         'plps': {1234567: [{'pk': 6142511,
            'module_id': ...,
            'code': '...',
            'name': '...',
            'learning_required': False,
            'learning_method': '...',
            'learning_completed': '...',
            'validated_membership_number': '...',
            'validated_name': '...'},
           ...],
          ...},
         'mandatory': {'GDPR':
          {'name': 'GDPR',
          'completed_date': datetime.datetime(...)},
          ...}}

    Raises:
        CompassNetworkError:
            For errors while executing the HTTP call

    """
    logger.debug(f"getting training tab for member number: {membership_number}")

    tree = _get_member_profile_tab(client, membership_number, "Training")
    rows = tree.xpath("//table[@id='tbl_p5_TrainModules']/tr")

    training_plps: TYPES_TRAINING_PLPS = {}
    training_roles = {}
    for row in rows:
        classes = set(row.classes)

        # Personal Learning Plan (PLP) data
        if "trPLP" in classes:
            plp_number, plp_data = _process_personal_learning_plan(row)
            training_plps[plp_number] = plp_data

        # Role data
        if "msTR" in classes:
            role_number, role_data = _process_role_data(row)
            training_roles[role_number] = role_data

    details = {"roles": training_roles, "plps": training_plps, "mandatory": _compile_ongoing_learning(training_plps, tree)}
    with validation_errors_logging(membership_number):
        return ci.MemberTrainingTab.parse_obj(details)


def _process_personal_learning_plan(plp: html.HtmlElement) -> tuple[int, list[TYPES_TRAINING_MODULE]]:
    """Parses a personal learning plan from a LXML row element containing data."""
    plp_data = []
    plp_table = plp.getchildren()[0].getchildren()[0]
    for module_row in plp_table:
        if module_row.get("class") != "msTR trMTMN":
            continue

        module_data: TYPES_TRAINING_MODULE = {}
        child_nodes = list(module_row)
        module_data["pk"] = int(module_row.get("data-pk"))
        module_data["module_id"] = int(child_nodes[0].get("id")[4:])
        matches = re.match(r"^([A-Z0-9]+) - (.+)$", child_nodes[0].text_content()).groups()  # type: ignore[union-attr]
        if matches:
            module_data["code"] = str(matches[0])
            module_data["name"] = matches[1]

        learning_required = child_nodes[1].text_content().lower()
        module_data["learning_required"] = "yes" in learning_required if learning_required else None
        module_data["learning_method"] = child_nodes[2].text_content() or None
        module_data["learning_completed"] = parse_date(child_nodes[3].text_content())
        module_data["learning_date"] = parse_date(child_nodes[3].text_content())

        validated_by_string = child_nodes[4].text_content()
        if validated_by_string:
            # Add empty item to prevent IndexError
            validated_by_data = validated_by_string.split(" ", 1) + [""]
            module_data["validated_membership_number"] = maybe_int(validated_by_data[0])
            module_data["validated_name"] = validated_by_data[1]
        module_data["validated_date"] = parse_date(child_nodes[5].text_content())

        plp_data.append(module_data)

    return int(plp_table.get("data-pk")), plp_data


def _process_role_data(role: html.HtmlElement) -> tuple[int, dict[str, Union[None, str, int, datetime.date]]]:
    """Parses a personal learning plan from a LXML row element containing data."""
    child_nodes = list(role)

    role_data: dict[str, Union[None, str, int, datetime.date]] = dict()
    role_number = int(role.get("data-ng_mrn"))
    role_data["role_number"] = role_number
    role_data["role_title"] = child_nodes[0].text_content()
    role_data["role_start"] = parse_date(child_nodes[1].text_content())
    status_with_review = child_nodes[2].text_content()
    # TODO for `Ending: blah` roles, should we store the ending date?
    if status_with_review.startswith("Full (Review Due: ") or status_with_review.startswith("Full (Ending: "):
        role_data["role_status"] = "Full"
        date_str = status_with_review.removeprefix("Full (Review Due: ").removeprefix("Full (Ending: ").rstrip(")")
        role_data["review_date"] = parse_date(date_str)
    else:
        role_data["role_status"] = status_with_review
        role_data["review_date"] = None
    role_data["location"] = child_nodes[3].text_content()
    training_advisor_string = child_nodes[4].text_content()
    if training_advisor_string:
        role_data["ta_data"] = training_advisor_string
        # Add empty item to prevent IndexError
        training_advisor_data = training_advisor_string.split(" ", 1) + [""]
        role_data["ta_number"] = maybe_int(training_advisor_data[0])
        role_data["ta_name"] = training_advisor_data[1]
    completion_string = child_nodes[5].text_content()
    if completion_string:
        role_data["completion"] = completion_string
        parts = completion_string.split(":")
        role_data["completion_type"] = parts[0].strip()
        role_data["completion_date"] = parse_date(parts[1].strip())
    role_data["wood_badge_number"] = child_nodes[5].get("id", "").removeprefix("WB_") or None

    return role_number, role_data


def _compile_ongoing_learning(training_plps: TYPES_TRAINING_PLPS, tree: html.HtmlElement) -> TYPES_TRAINING_OGL:
    """Compiles ongoing learning data.

    Uses PLP records for GDPR, and main OGL section for Safety, Safeguarding,
    and First Aid.

    Args:
        training_plps: Parsed PLP data.
        tree: LXML tree representing training tab

    Returns:
        A map of ogl type -> dates. Data returned here as native types (dicts),
        see ci.MemberMandatoryTraining for full model.

    """
    # Handle GDPR (Get latest GDPR date)
    training_ogl: TYPES_TRAINING_OGL = dict()
    gdpr_generator: Iterator[datetime.date] = (
        module["validated_date"]
        for plp in training_plps.values()
        for module in plp
        if module["code"] == "GDPR" and isinstance(module["validated_date"], datetime.date)
    )
    training_ogl["gdpr"] = dict(completed_date=next(reversed(sorted(gdpr_generator)), None))

    # Get main OGL - safety, safeguarding, first aid
    for ongoing_learning in tree.xpath("//tr[@data-ng_code]"):
        cell_text = {c.get("id", "<None>").split("_")[0]: c.text_content() for c in ongoing_learning}
        # TODO missing data-pk from list(cell)[0].tag == "input", and module names/codes. Are these important?
        training_ogl[mogl_modules[ongoing_learning.get("data-ng_code")]] = dict(
            completed_date=parse_date(cell_text.get("tdLastComplete", "")),
            renewal_date=parse_date(cell_text.get("tdRenewal", "")),
        )

    # Update training_ogl with missing mandatory ongoing learning types
    blank_module: TYPES_TRAINING_OGL_DATES = dict(completed_date=None, renewal_date=None)
    return {mogl_type: training_ogl.get(mogl_type, blank_module) for mogl_type in mogl_modules.values()}


@cache_hooks.cache_result(key=("awards", 1))
def get_awards_tab(client: Client, membership_number: int, /) -> list[ci.MemberAward]:
    """Returns data from Awards tab for a given member.

    Args:
        client: HTTP client
        membership_number: Membership Number to use

    Returns:
        A MemberAward object with corresponding data from the awards tab
        (keys will always be present):

        MemberAward(
            membership_number=...,
            type="...",
            location="...",
            date=datetime.date(...),
        )

    Raises:
        CompassNetworkError:
            For errors while executing the HTTP call

    """
    tree = _get_member_profile_tab(client, membership_number, "Awards")
    rows = tree.xpath("//table[@class='msAward']/tr")

    awards = []
    with validation_errors_logging(membership_number):
        for row in rows:
            award_props = row[1][0]  # Properties are stored as yet another sub-table
            awards.append(
                ci.MemberAward(
                    membership_number=membership_number,
                    type=award_props[0][1].text_content(),
                    location=award_props[1][1].text_content() or None,
                    date=parse_date(award_props[2][1].text_content() or ""),  # type  ignore[arg-type]
                )
            )
    return awards


@cache_hooks.cache_result(key=("disclosures", 1))
def get_disclosures_tab(client: Client, membership_number: int, /) -> list[ci.MemberDisclosure]:
    """Returns data from Disclosures tab for a given member.

    Args:
        client: HTTP client
        membership_number: Membership Number to use

    Returns:
        A MemberDisclosure object with corresponding data from the
        disclosures tab (keys will always be present):

        MemberDisclosure(
            membership_number=...,
            country="...",
            provider="...",
            type="...",
            number=...,
            issuer="...",
            issue_date=datetime.date(...),
            status="...",
            expiry_date=datetime.date(...),
        )

    Raises:
        CompassNetworkError:
            For errors while executing the HTTP call

    """
    tree = _get_member_profile_tab(client, membership_number, "Disclosures")
    rows = tree.xpath("//tbody/tr")

    disclosures = []
    with validation_errors_logging(membership_number):
        for row in rows:
            # Get children (cells in row)
            cells = list(row)

            disclosures.append(
                ci.MemberDisclosure(
                    membership_number=membership_number,
                    country=cells[0].text_content() or None,  # Country sometimes missing (Application Withdrawn)
                    provider=cells[1].text_content(),
                    type=cells[2].text_content(),
                    number=cells[3].text_content() or None,  # If Application Withdrawn, no disclosure number
                    issuer=cells[4].text_content() or None,
                    issue_date=parse_date(cells[5].text_content()),  # If Application Withdrawn, maybe no issue date
                    status=cells[6].text_content(),
                    expiry_date=parse_date(cells[7].text_content()),  # If Application Withdrawn, no expiry date
                )
            )
    return disclosures



from __future__ import annotations

import datetime
import time
from typing import Any, Optional, TYPE_CHECKING

from compass.core.logger import logger
from compass.core.settings import Settings
from compass.core.util.compass_helpers import compass_restify

if TYPE_CHECKING:
    from collections.abc import Mapping

    import requests

    from compass.core.util.client import Client  # pylint: disable=ungrouped-imports

TYPE_AUTH_IDS = tuple[int, int, str]


def jk_hash(client: Client, membership_number: int, role_number: int, jk: str) -> str:
    """Generate JK Hash needed by Compass."""
    # hash_code(f"{time.time() * 1000:.0f}")
    key_hash = f"{time.time() * 1000:.0f}{jk}{role_number}{membership_number}"  # JK, MRN & CN are all required.
    data = compass_restify({"pKeyHash": key_hash, "pCN": membership_number})
    logger.debug(f"Sending preflight data {datetime.datetime.now()}")
    client.post(f"{Settings.base_url}/System/Preflight", json=data)
    Settings.total_requests += 1
    return key_hash


def auth_header_get(
    auth_ids: TYPE_AUTH_IDS,
    client: Client,
    url: str,
    *,
    params: Optional[Mapping[str, Optional[str]]] = None,
    headers: Optional[Mapping[str, str]] = None,
    stream: Optional[bool] = None,
    **kwargs: Any,
) -> requests.Response:
    """Sends a HTTP GET request.

    Pass-through method to requests.sessions.Session.get, also adding to
    the counter of total requests sent by `compass.core`.

    Adds custom auth_header.py logic for certain Compass requests
    See Scouts.js -> $.ajaxSetup -> beforeSend for details (links below)
    https://github.com/the-scouts/compass-interface/blob/master/js/Scouts.js#L1122
    https://github.com/the-scouts/compass-interface/blob/master/js/Scouts.js#L1128-L1129

    Logic in source comments is auth_header.py logic is only added for **AJAX**
    GET calls matching the following logic:

    if method == "GET":
        if compass_props.master.sys.safe_json is True and "system/preflight" not in url:
            return True
        elif url.lower().replace(Settings.web_service_path.lower(), "").startswith("sto_check")
            return False
        return True

    Args:
        auth_ids:
            Three-tuple of Current authenticated user's membership number,
            active role number, and 'jk' (Ideas: Join Key??? SHA2-512 like)
        client: HTTP client
        url: Request URL
        params: Mapping to be sent in the query string for the request
        headers: Mapping of HTTP Headers
        stream: Whether to stream download the response content.
        kwargs: Optional arguments to session.get

    Returns:
        requests.Response object from executing the request

    Raises:
        CompassNetworkError

    """
    # pylint: disable=too-many-arguments
    # pylint complains that we have more than 5 arguments.
    membership_number, role_number, jk = auth_ids
    headers = dict(headers or {}) | {"Auth": jk_hash(client, membership_number, role_number, jk)}

    params = dict(params or {}) | {
        "x1": f"{membership_number}",
        "x2": f"{jk}",
        "x3": f"{role_number}",
    }

    return client.get(url, params=params, headers=headers, stream=stream, **kwargs)



from __future__ import annotations

import datetime
from typing import Optional, TYPE_CHECKING, Union

from lxml import html

import compass.core as ci
from compass.core.logger import logger
from compass.core.settings import Settings
from compass.core.util import cache_hooks
from compass.core.util.context_managers import validation_errors_logging
from compass.core.util.type_coercion import maybe_int
from compass.core.util.type_coercion import parse_date

if TYPE_CHECKING:
    from collections.abc import Iterator

    from compass.core.util.client import Client

renamed_modules = {
    "001": "module_01",
    "TRST": "trustee_intro",
    "002": "module_02",
    "003": "module_03",
    "004": "module_04",
    "GDPR": "GDPR",
    "SFTY": "safety",
    "SAFE": "safeguarding",
}
unset_vals = {"--- Not Selected ---", "--- No Items Available ---", "--- No Line Manager ---"}

module_names = {
    "Essential Information": "M01",
    "Trustee Introduction": "TRST",
    "Personal Learning Plan": "M02",
    "Tools for the Role (Section Leaders)": "M03",
    "Tools for the Role (Managers and Supporters)": "M04",
    "General Data Protection Regulations": "GDPR",
    "Safety Training": "SFTY",
    "Safeguarding Training": "SAFE",
}

references_codes = {
    "NC": "Not Complete",
    "NR": "Not Required",
    "RR": "References Requested",
    "S": "References Satisfactory",
    "U": "References Unsatisfactory",
}


# See getAppointment in PGS\Needle
@cache_hooks.cache_result(key=("role_detail", 1))
def get_roles_detail(client: Client, role_number: int, /) -> ci.MemberRolePopup:
    """Returns detailed data from a given role number.

    Args:
        client: HTTP client
        role_number: Role Number to use

    Returns:
        A MemberRolePopup model with the data from the role detail popup
        (keys will always be present):

        {'hierarchy': {'organisation': 'The Scout Association',
          'country': '...',
          'region': '...',
          'county': '...',
          'district': '...',
          'group': '...',
          'section': '...'},
         'details': {'role_number': ...,
          'organisation_level': '...',
          'birth_date': datetime.datetime(...),
          'membership_number': ...,
          'name': '...',
          'role_title': '...',
          'role_start': datetime.datetime(...),
          'role_status': '...',
          'line_manager_number': ...,
          'line_manager': '...',
          'ce_check': datetime.datetime(...),
          'disclosure_check': '...',
          'references': '...',
          'appointment_panel_approval': '...',
          'commissioner_approval': '...',
          'committee_approval': '...'},
         'getting_started': {...: {'name': '...',
           'validated': datetime.datetime(...),
           'validated_by': '...'},
           ...
          }}

    Raises:
        CompassNetworkError:
            For errors while executing the HTTP call

    """
    response = client.get(f"{Settings.base_url}/Popups/Profile/AssignNewRole.aspx?VIEW={role_number}")
    tree = html.fromstring(response.content)
    if tree.forms[0].action == "./ScoutsPortal.aspx?Invalid=Access":
        raise ci.CompassPermissionError(f"You do not have permission to the details of role {role_number}")

    inputs: dict[str, Union[html.InputElement, html.SelectElement]] = dict(tree.forms[0].inputs)
    fields: dict[str, str] = {k: v.value for k, v in inputs.items() if v.value is not None}

    line_manager_number, line_manager_name = _extract_line_manager(inputs["ctl00$workarea$cbo_p2_linemaneger"])
    ce_check = fields.get("ctl00$workarea$txt_p2_cecheck", "")  # CE (Confidential Enquiry) Check
    disclosure_check, disclosure_date = _extract_disclosure_date(fields.get("ctl00$workarea$txt_p2_disclosure", ""))

    approval_values = {row[1][0].get("data-app_code"): row[1][0].get("data-db") for row in tree.xpath("//tr[@class='trProp']")}
    # row[1][0].get("title") gives title text, but this is not useful as it does not reflect latest changes,
    # but only who added the role to Compass.

    role_details: dict[str, Union[None, int, str, datetime.date]] = dict(
        role_number=role_number,
        # `organisation_level` is ignored, no corresponding field in MemberTrainingRole:
        organisation_level=fields.get("ctl00$workarea$cbo_p1_level"),
        birth_date=parse_date(inputs["ctl00$workarea$txt_p1_membername"].get("data-dob")),
        membership_number=int(fields.get("ctl00$workarea$txt_p1_memberno", 0)),
        # `name` is ignored, no corresponding field in MemberTrainingRole:
        name=fields.get("ctl00$workarea$txt_p1_membername", " ").split(" ", 1)[1],
        role_title=fields.get("ctl00$workarea$txt_p1_alt_title"),
        role_start=parse_date(fields.get("ctl00$workarea$txt_p1_startdate", "")),
        role_status=fields.get("ctl00$workarea$txt_p2_status"),
        line_manager_number=line_manager_number,
        line_manager=line_manager_name,
        review_date=parse_date(fields.get("ctl00$workarea$txt_p2_review", "")),
        ce_check=parse_date(ce_check) if ce_check != "Pending" else None,  # TODO if CE check date != current date then is valid
        disclosure_check=disclosure_check,
        disclosure_date=disclosure_date,
        references=references_codes.get(fields.get("ctl00$workarea$cbo_p2_referee_status", "")),
        appointment_panel_approval=approval_values.get("ROLPRP|AACA"),
        commissioner_approval=approval_values.get("ROLPRP|CAPR"),
        committee_approval=approval_values.get("ROLPRP|CCA"),
    )

    logger.debug(f"Processed details for role number: {role_number}.")
    # TODO data-ng_id?, data-rtrn_id?

    full_details = {
        "hierarchy": dict(_process_hierarchy(inputs)),
        "details": {k: v for k, v in role_details.items() if v is not None},  # Filter null values
        "getting_started": _process_getting_started(tree.xpath("//tr[@class='trTrain trTrainData']")),
    }

    with validation_errors_logging(role_number, name="Role Number"):
        return ci.MemberRolePopup.parse_obj(full_details)


def _extract_line_manager(line_manager_list: html.SelectElement) -> tuple[Optional[int], Optional[str]]:
    line_manager_el = next((op for op in line_manager_list if op.get("selected")), None)
    if line_manager_el is None:
        return None, None
    number = maybe_int(line_manager_el.get("value"))
    name = line_manager_el.text.strip()
    if name in unset_vals:
        return number, None
    return number, name


def _extract_disclosure_date(disclosure_status: str) -> tuple[Optional[str], Optional[datetime.date]]:
    """Return tuple of disclosure check status, disclosure date."""
    if disclosure_status.startswith("Disclosure Issued : "):
        return "Disclosure Issued", parse_date(disclosure_status.removeprefix("Disclosure Issued : "))
    return (disclosure_status or None), None


def _process_hierarchy(inputs: dict[str, html.HtmlElement]) -> Iterator[tuple[str, str]]:
    """Get all levels of the org hierarchy and select those that will have information."""
    # Get all inputs with location data
    for input_name, input_el in inputs.items():
        if "ctl00$workarea$cbo_p1_location" not in input_name:
            continue
        level_name = input_el.get("title").lower()
        level_value = input_el[0].text
        if level_value not in unset_vals:
            yield "county" if level_name == "county / area / scottish region / overseas branch" else level_name, level_value


def _process_getting_started(getting_started_modules: html.HtmlElement) -> dict[str, dict[str, Union[None, str, datetime.date]]]:
    """Process getting started modules."""
    modules_output = {}
    # Get all training modules and then extract the required modules to a dictionary
    for module in getting_started_modules:
        module_name = module[0][0].text.strip()
        if module_name in module_names:
            info = {
                # "name": module_names[module_name],  # short_name
                "validated": parse_date(module[2][0].value),  # Save module validation date
                "validated_by": module[1][1].get("value") or None,  # Save who validated the module
            }
            mod_code: str = module[2][0].get("data-ng_value")
            modules_output[renamed_modules[mod_code]] = info

    return modules_output



from __future__ import annotations

from typing import Iterable, Optional, TYPE_CHECKING, TypedDict, Union

import compass.core as ci
from compass.core._scrapers import hierarchy as scraper
from compass.core.logger import logger
from compass.core.schemas import hierarchy as schema
from compass.core.util import cache_hooks

if TYPE_CHECKING:
    from collections.abc import Iterator

    from compass.core.util.client import Client

    class _HierarchyState(TypedDict, total=False):
        unit_id: int
        name: Optional[str]
        # organisation: int  # Always 10000001; mildly redundant
        country: int
        region: int
        county: int
        district: int
        group: int

    _TYPES_LEVEL = tuple[Optional[ci.TYPES_UNIT_LEVELS], Optional[scraper.TYPES_ENDPOINT_LEVELS], scraper.TYPES_ENDPOINT_LEVELS]


# Holds unit level metadata. Keys are from `schema.TYPES_UNIT_LEVELS`, values
# are a three-tuple of child level, unit level endpoint, and unit  level
# section endpoint.
# Suppress PyCharm inspection (not great at literals)
# noinspection PyTypeChecker
_levels: dict[ci.TYPES_UNIT_LEVELS, _TYPES_LEVEL] = {
    # "Group": (None, None, "group_sections"),
    "District": ("Group", "groups", "district_sections"),
    "County": ("District", "districts", "county_sections"),
    "Region": ("County", "counties", "region_sections"),
    "Country": ("Region", "regions", "country_sections"),
    "Organisation": ("Country", "countries", "hq_sections"),
}


class Hierarchy:
    def __init__(self, session: ci.Logon):
        """Constructor for Hierarchy."""
        self.client: Client = session._client
        self.default_hierarchy: ci.HierarchyLevel = session.hierarchy

    # See recurseRetrieve in PGS\Needle
    @cache_hooks.cache_result(key=("hierarchy", 1), model_type=schema.UnitData)
    def unit_data(
        self,
        unit_id: Optional[int] = None,
        level: Optional[ci.TYPES_HIERARCHY_LEVELS] = None,
        use_default: bool = False,
        recurse_children: bool = True,
    ) -> ci.UnitData:
        """Gets all units at given level and below, including sections.

        Unit data can be specified as a pre-constructed model, by passing literals, or
        by signalling to use the data from the user's current role. If all three
        options are unset an exception is raised.

        There is a strict priority ordering as follows:
            1. pre-constructed pydantic model
            2. literals
            3. default data

        Args:
            unit_id: Compass unit ID to get data for
            level: Level for unit ID (Group, District, ...)
            use_default: Use current role's hierarchy level (overridden by explicit unit_id/level)
            recurse_children: Return all data down to group sections, or just immediate children

        Raises:
            CompassError:
                When no unit data information has been provided

        """
        if use_default:
            unit_meta = self.default_hierarchy
        elif unit_id is not None and level is not None:
            unit_meta = ci.HierarchyLevel(unit_id=unit_id, level=level)
        else:
            raise ci.CompassError("No level data specified! Either `use_default` or both `unit_id` and `level` must be set!")
        # Fetch the hierarchy
        return ci.UnitData.parse_obj(_get_descendants_level(self.client, unit_meta, recurse_children))

    def unique_members(
        self,
        unit_id: Optional[int] = None,
        level: Optional[ci.TYPES_HIERARCHY_LEVELS] = None,
        use_default: bool = False,
        recurse_children: bool = True,
    ) -> set[int]:
        """Get all unique members for a given level and its descendants.

        Unit data can be specified as a pre-constructed model, by passing literals, or
        by signalling to use the data from the user's current role. If all three
        options are unset an exception is raised.

        There is a strict priority ordering as follows:
            1. pre-constructed pydantic model
            2. literals
            3. default data

        Returns:
            A set of unique member numbers within the given unit.

        Raises:
            CompassError:
                When no unit data information has been provided

        """
        # get tree of all units
        hierarchy_dict = self.unit_data(unit_id, level, use_default, recurse_children)

        # flatten tree
        flat_hierarchy = _flatten_hierarchy(hierarchy_dict)

        # generator for compass unit IDs
        compass_ids = (unit["unit_id"] for unit in flat_hierarchy)

        # get members from the list of IDs
        seen: set[int] = set()  # Unit ID deduplication
        add = seen.add
        unit_member_lists = (self.unit_members(sub_id) for sub_id in compass_ids if (sub_id not in seen and not add(sub_id)))

        # return a set of membership numbers
        return {members.contact_number for unit_member_list in unit_member_lists for members in unit_member_list}

    def units_members(self, unit_ids: Iterable[int]) -> Iterator[ci.HierarchyUnitMembers]:
        """Fetch all members from an iterable of unit IDs."""
        seen = set()  # Unit ID cache
        for unit_id in unit_ids:
            if unit_id in seen:
                continue
            seen.add(unit_id)  # store already fetched unit IDs
            yield ci.HierarchyUnitMembers(unit_id=unit_id, members=self.unit_members(unit_id))

    def unit_members(self, unit_id: int) -> list[ci.HierarchyMember]:
        logger.debug(f"Getting members for {unit_id}")
        return scraper.get_members_with_roles_in_unit(self.client, unit_id)


def _get_descendants_level(client: Client, unit_meta: ci.HierarchyLevel, recurse_children: bool) -> dict[str, object]:
    # short-circuit if level type is a section hierarchy type
    if unit_meta.level not in _levels:
        return {"unit_id": unit_meta.unit_id, "level": unit_meta.level, "child": [], "sections": []}
    if recurse_children:
        return _get_descendants_recursive(client, unit_meta.unit_id, unit_meta.level)  # type: ignore[arg-type]
    return _get_descendants_immediate(client, unit_meta.unit_id, unit_meta.level)  # type: ignore[arg-type]


# See recurseRetrieve in PGS\Needle
def _get_descendants_recursive(client: Client, unit_id: int, level_name: ci.TYPES_UNIT_LEVELS, /) -> dict[str, object]:
    """Recursively get all children from given unit ID and level."""
    logger.debug(f"getting data for unit {unit_id}")

    # All to handle as Group doesn't have grand-children
    unit_data: dict[str, object] = {"unit_id": unit_id, "level": level_name}

    # Do child units exist? (i.e. is this level != group)
    child_level, endpoint_children, endpoint_sections = _levels[level_name]
    if endpoint_children and child_level:
        children = scraper.get_units_from_hierarchy(client, unit_id, endpoint_children)
        # extend children with grandchildren
        unit_data["child"] = [child.__dict__ | _get_descendants_recursive(client, child.unit_id, child_level) for child in children]
    unit_data["sections"] = scraper.get_units_from_hierarchy(client, unit_id, endpoint_sections)

    return unit_data


def _get_descendants_immediate(client: Client, unit_id: int, level_name: ci.TYPES_UNIT_LEVELS, /) -> dict[str, object]:
    """Recursively get all children from given unit ID and level."""
    logger.debug(f"getting data for unit {unit_id}")

    # All to handle as Group doesn't have grand-children
    unit_data: dict[str, object] = {"unit_id": unit_id, "level": level_name}

    # Do child units exist? (i.e. is this level != group)
    child_level_name, endpoint_children, endpoint_sections = _levels[level_name]
    if endpoint_children:
        blank_descendant_data: dict[str, object] = {"level": child_level_name, "child": None, "sections": []}
        children = scraper.get_units_from_hierarchy(client, unit_id, endpoint_children)
        unit_data["child"] = [child.__dict__ | blank_descendant_data for child in children]
    unit_data["sections"] = scraper.get_units_from_hierarchy(client, unit_id, endpoint_sections)
    return unit_data


def _flatten_hierarchy(hierarchy_dict: ci.UnitData) -> Iterator[_HierarchyState]:  # noqa: D417 (hanging indent)
    """Flattens a hierarchy tree / graph to a flat sequence of mappings.

    Args:
        hierarchy_dict:
            The current object to be flattened. The user will pass in a `ci.UnitData`
            object, whilst all recursion will be on `ci.DescendantData` objects.

    """
    # This args style is allowed, but not yet (2021-03-20) implemented in PyDocStyle, so D417 disabled above.
    # https://github.com/PyCQA/pydocstyle/issues/449
    blank_state: _HierarchyState = {}
    return _flatten(hierarchy_dict, blank_state)


def _flatten(d: Union[ci.UnitData, ci.DescendantData], hierarchy_state: _HierarchyState) -> Iterator[_HierarchyState]:
    """Generator expresion to recursively flatten hierarchy."""
    unit_id = d.unit_id
    level_data = hierarchy_state | {d.level.lower(): unit_id}  # type: ignore[operator]
    yield {"unit_id": unit_id, "name": d.name, "section": False} | level_data  # type: ignore[misc]
    for child in d.child or []:
        yield from _flatten(child, level_data)  # type: ignore[arg-type]
    for section in d.sections:
        yield {"unit_id": section.unit_id, "name": section.name, "section": True} | level_data  # type: ignore[misc]



from __future__ import annotations

from typing import Literal, Optional, Union

import pydantic

TYPES_SECTION = Literal[
    "EY Pilot",
    "Beavers",
    "Cubs",
    "Scouts",
    "Explorers",
    "Network",
    "ASU",
    "Other",
]
TYPES_UNIT_LEVELS = Literal["Group", "District", "County", "Region", "Country", "Organisation"]
TYPES_SECTION_LEVELS = Literal[
    "Group Section",
    "District Section",
    "County Section",
    "Regional Section",
    "Country Section",
    "Organisation Section",
]
TYPES_HIERARCHY_LEVELS = Union[TYPES_UNIT_LEVELS, TYPES_SECTION_LEVELS]


class HierarchyBase(pydantic.BaseModel):
    unit_id: int


class HierarchyUnit(HierarchyBase):
    name: str


class HierarchySection(HierarchyUnit):
    # section details (only if a section!)
    section_type: Optional[TYPES_SECTION]


class HierarchyLevel(HierarchyBase):
    level: TYPES_HIERARCHY_LEVELS


class UnitData(HierarchyBase):
    name: Literal[None] = None
    level: TYPES_UNIT_LEVELS  # can't be a section
    child: Optional[list[DescendantData]]  # NOTE: deliberate recursive/forward reference here!
    sections: list[HierarchySection]


class DescendantData(UnitData, HierarchyUnit):
    name: str  # type: ignore[assignment]  # override UnitData name=None


UnitData.update_forward_refs()  # NOTE: updating recursive/forward reference here!
DescendantData.update_forward_refs()  # NOTE: updating recursive/forward reference here!


class HierarchyMemberID(pydantic.BaseModel):
    contact_number: int


class HierarchyMember(HierarchyMemberID):
    name: Optional[str]
    role: Optional[str]


class HierarchyUnitMembers(HierarchyBase):
    members: list[HierarchyMember]



from __future__ import annotations

import datetime
import json
from typing import Optional, TYPE_CHECKING

import compass.core as ci
from compass.core.settings import Settings
from compass.core.util.context_managers import validation_errors_logging
from compass.core.util.type_coercion import parse_date

if TYPE_CHECKING:
    from compass.core.util.client import Client

# /Contact/Avatar
# /Contact/ArchiveAlert
# /Contact/Alerts

lookup_nationality = {
    "AAAA": "British",
    "AAHL": "English",
    "AAAE": "Irish",
    "AAAB": "Northern Irish",
    "AAAC": "Scottish",
    "AAAD": "Welsh",
}

lookup_ethnicity = {
    "WHTE": "1.English/Welsh/Scottish/Northern Irish/British",
    "WIRH": "2.Irish",
    "WTRV": "3.Gypsy or Irish Traveller",
    "WOTH": "4.Any other White background",
    "MWBC": "5.White and Black Caribbean",
    "MWBA": "6.White and Black African",
    "MWAN": "7.White and Asian",
    "MOTH": "8.Any other mixed or Multiple ethic group",
    "AIND": "9.Indian",
    "APKN": "10.Pakistani",
    "ABGD": "11.Bangladeshi",
    "ACHN": "12.Chinese",
    "AOTH": "13.Any other Asian Background",
    "BAFR": "14.African",
    "BCRB": "15.Caribbean",
    "BOTH": "16.Any other Black/African/Caribbean background",
    "ARAB": "17.Arab",
    "OTHR": "18.Other",
    "PREF": "19.Prefer not to say",
}

lookup_religion = {
    "BUDD": "Buddhist",
    "CRST": "Christian (including all Christian denominations)",
    "HNDU": "Hindu",
    "JEW": "Jewish",
    "MSLM": "Muslim",
    "SIKH": "Sikh",
    "OTHR": "Any other religion (please specify)",
    "NONE": "No religion",
    "PREF": "Prefer not to say",
}

lookup_occupation = {
    "1": "Employed",
    "3": "Long term sick or disabled",
    "2": "Looking after home of family",
    "4": "Other",
    "12": "Student",
    "13": "Retired (whether receiving a pension or not)",
    "14": "Unemployed",
}


def get_contact_profile(client: Client, membership_number: int, /) -> ci.MemberDetails:
    request_json = {"Source": "ADULT", "ContactNumber": f"{membership_number}"}
    response = client.post(f"{Settings.base_url}/Contact/Profile", json=request_json)
    # yes, double loading the JSON is deliberate. The serialisation from Compass is insane...
    data = json.loads(json.loads(response.content.decode("utf-8")))

    if "PersonData" not in data:
        raise RuntimeError("!!!")

    person_data = data["PersonData"]

    details = {
        "membership_number": membership_number,
        "forenames": person_data["Forenames"],
        "surname": person_data["Surname"],
        "name": f"""{person_data["Forenames"]} {person_data["Surname"]}""",  # Full Name
        "known_as": person_data["KnownAs"],
        "join_date": parse_date(person_data["JoinDate"]),
        "sex": "Male" if person_data["Gender"] == "M" else "Female",
        "address": _process_address(person_data["Addresses"]),
        "main_phone": _process_phone_numbers(person_data["PhoneNumbers"]),
        "main_email": _process_email(person_data["EmailAddresses"]),
        "birth_date": parse_date(person_data["DOB"]),
        "nationality": lookup_nationality.get(person_data["Nationality"]["LookupValue"], "Other"),
        "ethnicity": lookup_ethnicity.get(person_data["Ethnicity"]["LookupValue"], ""),
        "religion": lookup_religion.get(person_data["Religion"]["LookupValue"], ""),
        "religion_detail": person_data["Religion"]["Details"],
        "occupation": lookup_occupation.get(person_data["Occupation"]["LookupValue"], ""),
        "occupation_detail": person_data["Occupation"]["Details"],
        "disabilities": _process_misc_sections(person_data["Disabilities"]),
        "qualifications": _process_misc_sections(person_data["Qualifications"]),
        "hobbies": _process_misc_sections(person_data["Hobbies"]),
    }
    # Filter out keys with no value.
    details = {k: v for k, v in details.items() if v}

    with validation_errors_logging(membership_number):
        return ci.MemberDetails.parse_obj(details)


def _process_address(addresses_props: list[dict[str, str]]) -> ci.AddressData:
    if not addresses_props:
        return {"unparsed_address": None, "country": None, "postcode": None, "county": None, "town": None, "street": None}
    address_props = addresses_props[0]
    return {
        "unparsed_address": address_props["Address"],
        "country": address_props["Country"],
        "postcode": address_props["Postcode"],
        "county": address_props["County"],
        "town": address_props["Town"],
        "street": f"""{address_props["Line1"]}, {address_props["Line2"]}, {address_props["Line3"]}""",
    }


def _process_phone_numbers(phone_numbers: list[dict[str, str]]) -> str:
    """Return a main phone number from a list of phone number properties.

    Each phone number record is as follows:
    {
        "CommunicationNumber": ...,
        "Type": "...",
        "Number": "...",
        "IsMain": bool
    }

    Args:
        phone_numbers: list of phone number records

    Todo:
        We could use the number type properties?

    Returns: main phone number

    """
    numbers = [""]
    for number_props in phone_numbers:
        if number_props["IsMain"]:
            return number_props["Number"]
        numbers.append(number_props["Number"])
    return numbers[0]


def _process_email(email_properties: list[dict[str, str]]) -> str:
    """Return a main phone number from a list of phone number properties.

    Each email record is as follows:
    {
        "CommunicationNumber": ...,
        "Type": "...",
        "EmailAddress": "...",
        "IsMain": bool
    }

    Args:
        email_properties: list of email records

    Todo:
        We could use the number type properties?

    Returns: main email address

    """
    emails = [""]
    for email_props in email_properties:
        if email_props["IsMain"]:
            return email_props["EmailAddress"]
        emails.append(email_props["EmailAddress"])
    return emails[0]


def _process_misc_sections(entries: list[str]) -> dict[str, str]:
    if not entries:
        return {}
    out = {}
    for entry in entries:
        field, _, optional_detail = entry.partition(" - ")
        out[field] = optional_detail
    return out


def get_contact_roles(client: Client, membership_number: int, /) -> ci.MemberRolesCollection:
    response = client.post(f"{Settings.base_url}/Contact/Roles", json={"ContactNumber": f"{membership_number}"})
    data = json.loads(response.content.decode("utf-8"))

    roles_data = {}
    for role_dict in data:
        role_details = ci.MemberRoleCore(
            role_number=role_dict["member_role_number"],
            membership_number=membership_number,
            role_title=role_dict["Role_Desc"],
            role_class=role_dict["class_desc"],
            role_type=role_dict["base_role_description"],
            location_id=role_dict["organisation_number"],
            location_name=role_dict["location"].strip(),
            role_start=_parse_iso_date(role_dict["start_date"]),  # type  ignore[arg-type]
            role_end=role_dict["end_Date"],  # None  # this API seems to only return open roles
            role_status=role_dict["status_desc"].split(" (")[0],  # TODO text processing
            review_date=_parse_iso_date(role_dict["review_date"]),
            can_view_details=False,  # TODO implement this
        )
        # status_code = role_dict["status"]  # TODO work out code -> status map ("A" == "Full"?)

        roles_data[role_details.role_number] = role_details

    primary_role = data[0]["member_role_number"] if data else None
    # We can't reliably calculate membership duration as only full roles seem
    # to be returned by the API here. So use magic value of -1 to signal this
    return ci.MemberRolesCollection(roles=roles_data, membership_duration=-1.0, primary_role=primary_role)


def _parse_iso_date(date_str: Optional[str]) -> Optional[datetime.date]:
    """Parses ISO 8601 datetime. Values might be None from the API."""
    if date_str is None:
        return None
    return datetime.datetime.fromisoformat(date_str).date()



from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from compass.core._scrapers import member_profile
from compass.core._scrapers import role_detail

if TYPE_CHECKING:
    import compass.core as ci

# SCRAPER CLASS - 1-1 mapping with compass to minimise calls
# MAIN CLASS - object/properties focused, with abstractions of actual calls


class People:
    def __init__(self, session: ci.Logon):
        """Constructor for People."""
        self.client = session._client

    def personal(self, membership_number: int) -> ci.MemberDetails:
        """Gets personal tab data for a given member.

        Args:
            membership_number: Membership Number to use

        Returns:
            A MemberDetails object containing all data.

        Raises:
            CompassNetworkError:
                For errors while executing the HTTP call
            CompassPermissionError:
                If the current user does not have permission to view personal
                data for the requested member.

        """
        return member_profile.get_personal_tab(self.client, membership_number)

    # See getRole in PGS\Needle
    def roles(
        self,
        membership_number: int,
        *,
        only_volunteer_roles: bool = True,
        only_active: bool = False,
    ) -> ci.MemberRolesCollection:
        """Gets the data from the Role tab in Compass for the given member.

        Parses the data to a common format, and removes Occasional Helper, PVG,
        Network, Council and Staff roles by default.

        Args:
            membership_number: Membership Number to use
            only_volunteer_roles: If True, only volunteer roles are returned,
                and Helper, Council, Network, etc are dropped.
            only_active: If True, inactive roles (Closed, Cancelled) are not
                returned.

        Returns:
            A MemberRolesDict object containing all data.

        Raises:
            CompassNetworkError:
                For errors while executing the HTTP call
            CompassPermissionError:
                If the current user does not have permission to view roles
                data for the requested member.

        """
        roles_data = member_profile.get_roles_tab(self.client, membership_number, only_volunteer_roles)
        if only_active is False:
            return roles_data
        # Role status filter
        status_blacklist = {"Closed", "Cancelled"}  # Inactive roles inferred from `ROLE_HIDEME` css class
        filtered_roles = {number: model for number, model in roles_data.roles.items() if model.role_status not in status_blacklist}
        # if role list hasn't changed return original model
        if filtered_roles.keys() == roles_data.roles.keys():
            return roles_data
        # don't mutate cached model
        return roles_data.__class__.parse_obj(roles_data.__dict__ | {"roles": filtered_roles})

    def role_detail(self, role_number: int) -> ci.MemberRolePopup:
        """Get detailed information for specified role.

         Make sure to check `MemberRoleCore.can_view_details` on each role, as
         you may not have permissions to view certain roles (if they are
         outside of your hierarchy, for example).

         Args:
             role_number: Role to query

        Returns:
            MemberRolePopup object with detail on the role

        Raises:
            CompassNetworkError:
                For errors while executing the HTTP call
            CompassPermissionError:
                If the current user does not have permission to view role data

        """
        return role_detail.get_roles_detail(self.client, role_number)

    def permits(self, membership_number: int) -> list[ci.MemberPermit]:
        """Gets permits tab data for a given member.

        Args:
            membership_number: Membership Number to use

        Returns:
            CompassNetworkError:
                For errors while executing the HTTP call
            A MemberPermitsList object containing all data.

        """
        return member_profile.get_permits_tab(self.client, membership_number)

    def training(self, membership_number: int) -> ci.MemberTrainingTab:
        """Gets training tab data for a given member.

        Args:
            membership_number: Membership Number to use

        Returns:
            CompassNetworkError:
                For errors while executing the HTTP call
            A MemberTrainingTab object containing all data.

        """
        return member_profile.get_training_tab(self.client, membership_number)

    def awards(self, membership_number: int) -> list[ci.MemberAward]:
        """Gets awards tab data for a given member.

        Args:
            membership_number: Membership Number to use

        Returns:
            A MemberAward object containing all data.

        Raises:
            CompassNetworkError:
                For errors while executing the HTTP call
            CompassPermissionError:
                If the current user does not have permission to view awards
                data for the requested member.

        """
        return member_profile.get_awards_tab(self.client, membership_number)

    def disclosures(self, membership_number: int) -> list[ci.MemberDisclosure]:
        """Gets disclosures tab data for a given member.

        Args:
            membership_number: Membership Number to use

        Returns:
            A list of MemberDisclosure objects containing all data.

        Raises:
            CompassNetworkError:
                For errors while executing the HTTP call
            CompassPermissionError:
                If the current user does not have permission to view
                disclosure data for the requested member.

        """
        return member_profile.get_disclosures_tab(self.client, membership_number)

    # Convenience methods:

    def adult_service_length(self, membership_number: int) -> float:
        """Returns length of service in qualifying roles in years.

        Excludes non-volunteer roles (e.g. Occasional Helper, PVG, Network),
        and returns total membership duration in fractional years. Correctly
        accounts for role overlap and gaps in service, as well as role statuses

        Args:
            membership_number: Membership Number to use

        Returns:
            Length of service in fractional years

        Raises:
            CompassNetworkError:
                For errors while executing the HTTP call
            CompassPermissionError:
                If the current user does not have permission to view roles
                data for the requested member.

        """
        return member_profile.get_roles_tab(self.client, membership_number).membership_duration

    def ongoing_learning(self, membership_number: int) -> ci.MemberMandatoryTraining:
        """Gets ongoing learning data for a given member.

        Args:
            membership_number: Membership Number to use

        Returns:
            CompassNetworkError:
                For errors while executing the HTTP call
            A MemberMOGLList object containing all data.

        """
        return member_profile.get_training_tab(self.client, membership_number).mandatory

    def latest_disclosure(self, membership_number: int) -> Optional[ci.MemberDisclosure]:
        """Gets latest disclosure for a given member.

        Latest is defined as the expiry date that is furthest away.

        Args:
            membership_number: Membership Number to use

        Returns:
            A MemberDisclosure objects containing disclosure data.

        Raises:
            CompassNetworkError:
                For errors while executing the HTTP call
            CompassPermissionError:
                If the current user does not have permission to view
                disclosure data for the requested member.

        """
        disclosures = self.disclosures(membership_number)
        date_map = {disc.expiry_date: disc for disc in disclosures if disc.expiry_date}
        return date_map[max(date_map)] if date_map else None



from __future__ import annotations

import json
from typing import get_args, Literal, TYPE_CHECKING

from lxml import html

import compass.core as ci
from compass.core.settings import Settings
from compass.core.util.compass_helpers import compass_restify

if TYPE_CHECKING:
    from compass.core.util.client import Client

# TYPES_ENDPOINT_LEVELS values are meaningful values as they become the API endpoint paths
TYPES_ENDPOINT_LEVELS = Literal[
    "countries",
    "hq_sections",
    "regions",
    "country_sections",
    "counties",
    "region_sections",
    "districts",
    "county_sections",
    "groups",
    "district_sections",
    "group_sections",
]
endpoints = {i: f"/{i.replace('_', '/')}" for i in get_args(TYPES_ENDPOINT_LEVELS)}
section_type_map = {
    "Early Years Pilot": "EY Pilot",
    "Beavers": "Beavers",
    "Beaver Scout": "Beavers",
    "Cub Scout": "Cubs",
    "Scout": "Scouts",
    "Explorer Scouts": "Explorers",
    "Scout Network": "Network",
    "Scout Active Support": "ASU",
    "All": "Other",
    "Other": "Other",
}


# see CompassClient::retrieveLevel or retrieveSections in PGS\Needle php
def get_units_from_hierarchy(client: Client, parent_unit: int, level: TYPES_ENDPOINT_LEVELS) -> list[ci.HierarchyUnit]:
    """Get all children of a given unit.

    If LiveData=Y is passed, the resulting JSON additionally contains:
        - (duplicated) parent id
        - the unit address
        - number of members
        - section type details, if requesting sections data

    Args:
        client: HTTP client
        parent_unit: The unit ID to get descendants from
        level: string org type, used for selecting API endpoint

    Returns:
        List of HierarchySection or HierarchyUnit models. If the current
        user cannot access the hierarchy unit, an empty list is returned.

        Returned lists have uniform types (i.e. all elements are of type
        HierarchySection or of type HierarchyUnit, they cannot mix)

        E.g.:
        [
            HierarchyUnit(
                unit_id=...,
                name='...',
                parent_id=...,
                status='...',
                address='...',
                member_count=...
            ),
            ...
        ]

    Raises:
        CompassNetworkError:
            For errors while executing the HTTP call

    """
    # Get API endpoint from level
    level_endpoint = endpoints[level]
    # Are we requesting sections here?
    is_sections = "/sections" in level_endpoint
    model_class = ci.HierarchySection if is_sections else ci.HierarchyUnit  # chose right model to use

    # LiveData causes a ~75x slowdown for non-section data, and we only use LiveData for sections anyway
    json_data = ({"LiveData": "Y"} if is_sections else {}) | {"ParentID": f"{parent_unit}"}
    result = client.post(f"{Settings.base_url}/hierarchy{level_endpoint}", json=json_data)
    result_json = result.json()

    # Handle unauthorised access
    if result_json == {"Message": "Authorization has been denied for this request."}:
        raise ci.CompassPermissionError(f"You do not have permission for unit ID:{parent_unit}! E:{level} S:{is_sections}")

    result_units = []
    for unit_dict in result_json:
        parsed = {
            "unit_id": int(unit_dict["Value"]),
            "name": unit_dict["Description"].strip(),
        }
        if unit_dict["Tag"]:
            tag = json.loads(unit_dict["Tag"])[0]
            # Only include section_type if there is section type data
            if "SectionTypeDesc" in tag or "SectionTypeDesc1" in tag:
                section_type = tag.get("SectionTypeDesc") or tag.get("SectionTypeDesc1") or "MISSING"
                parsed["section_type"] = section_type_map.get(section_type, section_type)

        result_units.append(model_class(**parsed))
    return result_units


def get_members_with_roles_in_unit(
    client: Client, unit_number: int, include_name: bool = False, include_primary_role: bool = False
) -> list[ci.HierarchyMember]:
    """Get details of members with roles in a given unit.

    Keys within the member_data JSON are (as at 13/01/220):
     - contact_number (membership number)
     - name (member's name)
     - visibility_status (this is meaningless as we can only see Y people)
     - address (this doesn't reliably give us postcode and is a lot of data)
     - role (This is Primary role and so only sometimes useful)

    Args:
        client: HTTP client
        unit_number: Compass unit number
        include_name: include member name in returned data
        include_primary_role: include primary role in returned data

    Returns:
        A list of member records. Keys are included through args

        E.g.:
        [
            {"contact_number": ..., ...},
            ...
        ]

    Raises:
        CompassNetworkError:
            For errors while executing the HTTP call
        CompassError:
            If Compass reports that the search was invalid

    """
    keys_to_keep = tuple(filter(None, ("contact_number", "name" * include_name, "role" * include_primary_role)))
    # Construct request data

    # It seems like the time UID value can be constant -- keeping old code in case something breaks
    # dt = datetime.datetime.now()
    # time_uid = f"{dt.hour}{dt.minute}{dt.microsecond // 1000}"
    time_uid = str(12_34_567)
    data = {"SearchType": "HIERARCHY", "OrganisationNumber": unit_number, "UI": time_uid}

    # Execute search
    # JSON data MUST be in the rather odd format of {"Key": key, "Value": value} for each (key, value) pair
    client.post(f"{Settings.base_url}/Search/Members", json=compass_restify(data))

    # Fetch results from Compass
    search_results = client.get(f"{Settings.base_url}/SearchResults.aspx")

    # Gets the compass form from the returned document
    form = html.fromstring(search_results.content).forms[0]
    del search_results

    # If the search hasn't worked the form returns an InvalidSearchError
    if form.action == "./ScoutsPortal.aspx?Invalid=SearchError":
        raise ci.CompassError("Invalid Search")

    # Get the encoded JSON data from the HTML
    member_data_string = form.fields["ctl00$plInnerPanel_head$txt_h_Data"] or "[]"
    del form

    # parse the data and return it as a usable Python object (list)
    member_data = json.loads(member_data_string)
    return [ci.HierarchyMember(**{key: member[key] for key in keys_to_keep}) for member in member_data]



# from __future__ import annotations
#
# import asyncio
# import contextlib
# import functools
# import threading
# from typing import Any, TYPE_CHECKING
#
# if TYPE_CHECKING:
#     from collections.abc import Callable
#
#
# class PeriodicTimer:
#     def __init__(self, interval: float, callback: Callable[..., Any]):
#         """Constructor for PeriodicTimer."""
#         self.interval = interval
#
#         @functools.wraps(callback)
#         def wrapper(*args: Any, **kwargs: Any) -> None:
#             result = callback(*args, **kwargs)
#             if result is not None:
#                 self.thread = threading.Timer(self.interval, self.callback)
#                 self.thread.start()
#
#         self.callback = wrapper
#         self.thread: threading.Timer = threading.Timer(0.0, self.callback)
#
#     def start(self) -> "PeriodicTimer":
#         self.thread.start()
#         return self
#
#     def cancel(self) -> "PeriodicTimer":
#         self.thread.cancel()
#         return self
#
#
# def set_interval(interval: int):
#     def decorator(function: Callable):
#         def wrapper(*args, **kwargs):
#             stopped = threading.Event()
#
#             def loop():  # executed in another thread
#                 while not stopped.wait(interval):  # until stopped
#                     function(*args, **kwargs)
#
#             t = threading.Thread(target=loop)
#             t.daemon = True  # stop if the program exits
#             t.start()
#             return stopped
#         return wrapper
#     return decorator
#
#
# async def periodic(n):
#     print('periodic')  # run immediately
#     while await asyncio.sleep(n, result=True):
#         print('periodic')
#
# loop = asyncio.get_event_loop()
# task = loop.create_task(periodic(0.5))
# loop.call_later(5, task.cancel)
# with contextlib.suppress(asyncio.CancelledError):
#     loop.run_until_complete(task)
#
#
# # wrapper:
# def periodic(period: int):
#     def scheduler(function: Callable):
#         async def wrapper(*args, **kwargs):
#             asyncio.create_task(function(*args, **kwargs))  # run immediately
#             while await asyncio.sleep(period, result=True):
#                 asyncio.create_task(function(*args, **kwargs))
#         return wrapper
#     return scheduler



from __future__ import annotations

import logging
from typing import Optional, TYPE_CHECKING

from compass.core.settings import Settings

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger("compass_interface")


def enable_debug_logging(*, log_file: Optional[Path] = None) -> None:
    import sys  # Disable check; only want to import sys if needed. pylint: disable=import-outside-toplevel

    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter(fmt="{asctime}.{msecs:03.0f} ({levelname}): {message}", datefmt="%Y-%m-%d %H:%M:%S", style="{")

    handlers = [logging.StreamHandler(stream=sys.stdout)]  # always include a default console handler
    if log_file is not None:
        handlers.append(logging.FileHandler(log_file.with_suffix(".log"), encoding="utf-8"))

    for handler in handlers:
        handler.setFormatter(formatter)
        logger.addHandler(handler)


if Settings.debug:
    enable_debug_logging(log_file=Settings.log_file)



class CompassError(Exception):
    pass


class CompassAuthenticationError(CompassError):
    pass


class CompassNetworkError(CompassError):
    pass


class CompassPermissionError(CompassError):
    pass


class CompassReportError(CompassError):
    pass


class CompassReportPermissionError(CompassReportError, CompassPermissionError):
    pass



import ctypes
import functools
from typing import Any


def hash_code(text: str) -> int:
    """Implements Java's hashCode in Python.

    Ref: https://stackoverflow.com/a/8831937
    """
    return functools.reduce(lambda code, char: ctypes.c_int32(31 * code + ord(char)).value, list(text), 0)


def compass_restify(data: dict[str, Any]) -> list[dict[str, str]]:
    """Format a dictionary of key-value pairs into the correct format for Compass.

    It seems that JSON data MUST be in the rather odd format of {"Key": key, "Value": value} for each (key, value) pair.
    """
    return [{"Key": f"{k}", "Value": f"{v}"} for k, v in data.items()]



import datetime
from typing import Any, Optional

months_lookup = {
    "january": 1,
    "jan": 1,
    "february": 2,
    "feb": 2,
    "march": 3,
    "mar": 3,
    "april": 4,
    "apr": 4,
    "may": 5,
    "june": 6,
    "jun": 6,
    "july": 7,
    "jul": 7,
    "august": 8,
    "aug": 8,
    "september": 9,
    "sep": 9,
    "october": 10,
    "oct": 10,
    "november": 11,
    "nov": 11,
    "december": 12,
    "dec": 12,
}


def maybe_int(value: Any) -> Optional[int]:
    """Casts value to int or None."""
    try:
        return int(value)
    except ValueError:
        return None


def parse_date(value: str) -> Optional[datetime.date]:
    """Custom date parsing function.

    datetime.datetime.strptime works, this is 10x quicker

    Two date formats:
     - %d %B %Y  e.g. 01 January 2000
     - %d %b %Y  e.g. 01 Jan 2000

     Args:
         value: date string to parse

    Returns:
        Parsed date object

    Raises:
        ValueError: If given string is not in a recognised date format

    """
    if not value:
        return None
    try:
        day, month, year = value.split(" ", 2)
        return datetime.date(int(year), months_lookup[month.lower()], int(day))
    except (KeyError, ValueError) as err:
        raise ValueError(f"Parsing string `{value}` into a date failed!") from err



import compass.core as ci
from compass.core._scrapers.reports import export_report
from compass.core._scrapers.reports import TYPES_REPORTS

__all__ = ("Reports", "TYPES_REPORTS")  # only needed whilst still no schema file for reports


class Reports:
    def __init__(self, session: ci.Logon):
        """Constructor for Reports."""
        self.auth_ids = session.membership_number, session.role_number, session._jk
        self.client = session._client

    def get_report(self, report_type: TYPES_REPORTS) -> bytes:
        """Exports report as CSV from Compass.

        Exporting a report is of course surprisingly complicated. The process
        has four major steps, as follows:

        1. Get a token for generating the report from the Compass backend. This
            also validates that the report exists and that the user is
            authenticated to access it.
        2. If successful in obtaining a report token, get the initial report
            page. The token is the (relative) URL here, and the report page
            contains further needed information, such as the export URL and
            location data for a full export.
            (Compass does not include all organisational units in reports by
            default, and to export all data for a given unit and downwards, we
            need to add in these missing/unset levels manually).
        3. We update report configuration data (sent as form data), and check
            that we are not in an error state.
        4. We extract the export URL, download the content and save to disk.

        Pitfalls to be aware of in this process include that:
        - Compass checks user-agent headers in some parts of the process
        - There is a ten (10) minute default soft-timeout, which may run out
            before a report download has finished
        - If a requested report is too large, Compass can simply give up, often
            with an `OutOfMemory` error or similar

        Returns:
            Report output, as a bytes-encoded object.

        Raises:
            CompassReportError:
                - If the user passes an invalid report type
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            CompassNetworkError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        return export_report(self.client, self.auth_ids, report_type, stream=False)



from __future__ import annotations

import datetime
from typing import Literal, Optional

import pydantic

TYPES_CRUD_STRING = pydantic.constr(regex=r"^[CRUD]{0,4}$")
TYPES_ORG_LEVELS = Literal["ORG", "ORST", "CNTR", "CNST", "REG", "RGST", "CNTY", "CTST", "DIST", "DTST", "SGRP", "SGST"]
TYPES_ROLE = tuple[str, str]


class CompassPropsBase(pydantic.BaseModel):
    class Config:  # noqa: D106 (missing docstring)
        allow_population_by_field_name = True
        extra = "allow"


class CompassPropsNav(CompassPropsBase):
    action: Optional[Literal["None"]] = pydantic.Field(None, alias="Action")
    start_no: Optional[int] = pydantic.Field(None, alias="StartNo")
    start_page: Optional[int] = pydantic.Field(None, alias="StartPage")


class CompassPropsPage(CompassPropsBase):
    use_cn: Optional[int] = pydantic.Field(None, alias="UseCN")
    hide_badges: Optional[bool] = pydantic.Field(None, alias="HideBadges")
    croc: Optional[Literal["OK"]] = pydantic.Field(None, alias="Croc")
    hide_nominations: Optional[bool] = pydantic.Field(None, alias="HideNominations")
    can_delete_ogl_hrs: Optional[bool] = pydantic.Field(None, alias="CanDeleteOGLHrs")
    fold_name: Optional[Literal["MP_"]] = pydantic.Field(None, alias="FoldName")


class CompassPropsCRUD(CompassPropsBase):
    mdis: Optional[TYPES_CRUD_STRING] = pydantic.Field(None, alias="MDIS")  # type: ignore[valid-type]
    roles: Optional[TYPES_CRUD_STRING] = pydantic.Field(None, alias="ROLES")  # type: ignore[valid-type]
    pemd: Optional[TYPES_CRUD_STRING] = pydantic.Field(None, alias="PEMD")  # type: ignore[valid-type]
    mmmd: Optional[TYPES_CRUD_STRING] = pydantic.Field(None, alias="MMMD")  # type: ignore[valid-type]
    mvid: Optional[TYPES_CRUD_STRING] = pydantic.Field(None, alias="MVID")  # type: ignore[valid-type]
    perm: Optional[TYPES_CRUD_STRING] = pydantic.Field(None, alias="PERM")  # type: ignore[valid-type]
    trn: Optional[TYPES_CRUD_STRING] = pydantic.Field(None, alias="TRN")  # type: ignore[valid-type]


class CompassPropsUser(CompassPropsBase):
    is_me: Optional[bool] = pydantic.Field(None, alias="IsMe")


class CompassPropsMasterSSO(CompassPropsBase):
    on: Optional[int] = pydantic.Field(None, alias="ON")


class CompassPropsMasterUser(CompassPropsBase):
    cn: Optional[int] = pydantic.Field(None, alias="CN")
    mrn: Optional[int] = pydantic.Field(None, alias="MRN")
    on: Optional[int] = pydantic.Field(None, alias="ON")
    lvl: Optional[TYPES_ORG_LEVELS] = pydantic.Field(None, alias="LVL")
    jk: Optional[str] = pydantic.Field(None, alias="JK")


class CompassPropsMasterConst(CompassPropsBase):
    wales: Optional[int] = pydantic.Field(None, alias="Wales")
    scotland: Optional[int] = pydantic.Field(None, alias="Scotland")
    over_seas: Optional[int] = pydantic.Field(None, alias="OverSeas")
    hq: Optional[int] = pydantic.Field(None, alias="HQ")


class CompassPropsMasterSys(CompassPropsBase):
    session_id: Optional[str] = pydantic.Field(None, alias="SessionID")
    safe_json: Optional[bool] = pydantic.Field(None, alias="SafeJSON")
    web_path: Optional[pydantic.HttpUrl] = pydantic.Field(None, alias="WebPath")
    text_size: Optional[int] = pydantic.Field(None, alias="TextSize")
    timout: Optional[int] = pydantic.Field(None, alias="Timout")
    rest: Optional[bool] = pydantic.Field(None, alias="REST")
    hard_time: Optional[datetime.time] = pydantic.Field(None, alias="HardTime")
    hard_expiry: Optional[str] = pydantic.Field(None, alias="HardExpiry")
    timeout_extension: Optional[int] = pydantic.Field(None, alias="TimeoutExtension")
    ping: Optional[int] = pydantic.Field(None, alias="Ping")


class CompassPropsMaster(CompassPropsBase):
    sso: CompassPropsMasterSSO = pydantic.Field(CompassPropsMasterSSO(), alias="SSO")
    user: CompassPropsMasterUser = pydantic.Field(CompassPropsMasterUser(), alias="User")
    const: CompassPropsMasterConst = pydantic.Field(CompassPropsMasterConst(), alias="Const")
    sys: CompassPropsMasterSys = pydantic.Field(CompassPropsMasterSys(), alias="Sys")


class CompassPropsConstSys(CompassPropsBase):
    sto: Optional[int] = pydantic.Field(None, alias="STO")
    sto_ask: Optional[int] = pydantic.Field(None, alias="STO_ASK")


class CompassPropsConst(CompassPropsBase):
    sys: CompassPropsConstSys = pydantic.Field(CompassPropsConstSys(), alias="Sys")


class CompassProps(CompassPropsBase):
    nav: CompassPropsNav = pydantic.Field(CompassPropsNav(), alias="Nav")
    page: CompassPropsPage = pydantic.Field(CompassPropsPage(), alias="Page")
    crud: CompassPropsCRUD = pydantic.Field(CompassPropsCRUD(), alias="CRUD")
    user: CompassPropsUser = pydantic.Field(CompassPropsUser(), alias="User")
    master: CompassPropsMaster = pydantic.Field(CompassPropsMaster(), alias="Master")
    const: CompassPropsConst = pydantic.Field(CompassPropsConst(), alias="Const")



from pathlib import Path
from typing import Optional

import pydantic


class _SettingsModel(pydantic.BaseSettings):
    # Network
    base_domain: str = "compass.scouts.org.uk"
    wcf_json_endpoint: str = "/JSon.svc"  # Windows communication foundation JSON service endpoint
    base_url: pydantic.HttpUrl = f"https://{base_domain}"  # type: ignore[assignment]
    web_service_path: pydantic.HttpUrl = base_url + wcf_json_endpoint  # type: ignore[assignment]
    # Requests
    total_requests: int = 0
    # Application
    org_number: int = 10000001
    date_format: str = "%d %B %Y"  # dd Month YYYY
    # Environment
    debug: bool = False
    log_file: Optional[Path] = None
    use_cache: bool = False
    validation_errors: bool = True

    class Config:
        case_sensitive = False  # this is the default, but mark for clarity.
        env_prefix = "CI_"  # env variables named `REDIS_HOST` etc


Settings = _SettingsModel()



from typing import Any

import requests

import compass.core as ci
from compass.core.settings import Settings


class Client(requests.Session):
    """Counts the number of requests sent."""

    def request(self, *args: Any, **kwargs: Any) -> requests.Response:
        Settings.total_requests += 1
        try:
            return super().request(*args, **kwargs)
        except requests.RequestException as err:
            raise ci.CompassNetworkError("Network Error!") from err



from __future__ import annotations

from pathlib import Path
import re
import time
from typing import Literal, TYPE_CHECKING

from lxml import html
import requests

import compass.core as ci
from compass.core.logger import logger
from compass.core.settings import Settings
from compass.core.util import auth_header
from compass.core.util import context_managers

if TYPE_CHECKING:
    from compass.core.util.auth_header import TYPE_AUTH_IDS
    from compass.core.util.client import Client

# TODO move to schema.reports if created
# TODO remove location from start, to keep list small
_report_types: dict[str, int] = {
    # group reports
    "Group Appointments Report": 59,
    # district reports
    "District Appointments Report": 50,
    "District Member Directory Report": 51,
    # "District Member Directory 18 To 25 Years": ,
    "District Permits Report": 70,
    "District Disclosure Report": 78,
    "District Training Report": 79,
    "District Awards Report": 94,
    "District Disclosure Management Report": 102,
    # county reports
    "County/Area/Region Appointments Report": 48,
    "County/Area/Region Member Directory Report": 49,
    "County/Area/Region Member Directory 18 To 25 Years": 53,
    "County/Area/Region Permits Report": 69,
    "County/Area/Region Disclosure Report": 77,
    "County/Area/Region Training Report": 80,
    "County/Area/Region Awards Report": 95,
    "County Disclosure Management Report": 101,
    # region reports
    "Region Member Directory": 37,
    "Region Appointments Report": 52,
    "Region Permit Report": 72,
    "Region Disclosure Report": 76,
    "Region Training Report": 84,
    "Region Disclosure Management Report": 100,
}
TYPES_REPORTS = Literal[
    # group
    "Group Appointments Report",
    # district
    "District Appointments Report",
    "District Member Directory Report",
    "District Permits Report",
    "District Disclosure Report",
    "District Training Report",
    "District Awards Report",
    "District Disclosure Management Report",
    # county
    "County/Area/Region Appointments Report",
    "County/Area/Region Member Directory Report",
    "County/Area/Region Member Directory 18 To 25 Years",
    "County/Area/Region Permits Report",
    "County/Area/Region Disclosure Report",
    "County/Area/Region Training Report",
    "County/Area/Region Awards Report",
    "County Disclosure Management Report",
    # region
    "Region Member Directory",
    "Region Appointments Report",
    "Region Permit Report",
    "Region Disclosure Report",
    "Region Training Report",
    "Region Disclosure Management Report",
]


def export_report(client: Client, auth_ids: TYPE_AUTH_IDS, report_type: TYPES_REPORTS, stream: bool = False) -> bytes:
    """Exports report as CSV from Compass.

    See `Reports.get_report` for an overview of the export process

    Returns:
        Report output, as a bytes-encoded object.

    Raises:
        CompassReportError:
            - If the user passes an invalid report type
            - If Compass returns a JSON error
            - If there is an error updating the form data
        CompassReportPermissionError:
            If the user does not have permission to run the report
        CompassNetworkError:
            If there is an error in the transport layer, or if Compass
            reports a HTTP 5XX status code

    """
    if report_type not in _report_types:
        types = [*_report_types.keys()]
        raise ci.CompassReportError(f"{report_type} is not a valid report type. Valid report types are {types}") from None
    report_number = _report_types[report_type]

    # Get token for report type & role running said report:
    run_report_url = _get_report_token(client, auth_ids, report_number)

    # Get initial reports page, for export URL and config:
    logger.info("Generating report")
    report_page = client.get(run_report_url).content

    # Update form data & set location selection:
    _update_form_data(client, report_page, run_report_url, report_number)

    # Export the report:
    logger.info("Exporting report")
    export_url = _extract_report_export_url(report_page.decode("UTF-8"))

    time_string = time.strftime("%Y-%m-%d %H-%M-%S")  # colons are illegal on windows
    filename = f"Compass Export - {report_type} - {time_string}.csv"
    csv_export = _download_report(client, export_url, streaming=stream, filename=filename)

    # start = time.time()
    # TODO TRAINING REPORT ETC.
    # # TODO REPORT BODY HAS KEEP ALIVE URL KeepAliveUrl
    # p = PeriodicTimer(15, lambda: self.report_keep_alive(self.session, report_page.text))
    # self.session.sto_thread.start()
    # p.start()
    # # ska_url = _report_keep_alive(self.session, report_page.text)
    # try:
    #     _download_report(self.session, f"{Settings.base_url}/{export_url_path}", export_url_params, filename, )  # ska_url
    # except (ConnectionResetError, requests.ConnectionError):
    #     logger.info(f"Stopped at {datetime.datetime.now()}")
    #     p.cancel()
    #     self.session.sto_thread.cancel()
    #     raise
    # logger.debug(f"Exporting took {time.time() - start}s")

    return csv_export


def _get_report_token(client: Client, auth_ids: TYPE_AUTH_IDS, report_number: int) -> str:
    params = {
        "pReportNumber": str(report_number),
        "pMemberRoleNumber": str(auth_ids[1]),  # auth IDs are membership number, role number, 'jk'
    }
    logger.debug("Getting report token")
    response = auth_header.auth_header_get(
        auth_ids,
        client,
        f"{Settings.web_service_path}/ReportToken",
        params=params,
    )
    _error_status(response)

    report_token_uri = response.json().get("d", "")
    if report_token_uri not in {"-1", "-2", "-3", "-4"}:
        return f"{Settings.base_url}/{report_token_uri}"
    if report_token_uri in {"-2", "-3"}:
        raise ci.CompassReportError("Report aborted: Report No Longer Available")
    if report_token_uri == "-4":  # nosec (false positive B105; not a hardcoded passwordstring)
        raise ci.CompassReportPermissionError("Report aborted: USER DOES NOT HAVE PERMISSION")
    raise ci.CompassReportError("Report aborted")


def _update_form_data(client: Client, report_page: bytes, run_report: str, report_number: int) -> None:
    # TODO add method to choose between exporting all data and just top-level
    tree = html.fromstring(report_page)

    # get relevant form data. TODO - do we need all items or just __VIEWSTATE?
    # form_data = {"__VIEWSTATE": next((el.value for el in tree.forms[0].iter("input") if el.name == "__VIEWSTATE"), "")}
    form_data = {el.name: el.value for el in tree.forms[0].inputs if el.get("type") not in {"checkbox", "image"}}

    # Appointments Reports
    if report_number == 52:
        form_data = _form_data_appointments(form_data, tree)

    # Compass does user-agent sniffing in reports!!! This does seem to be the
    # only place that *requires* a Mozilla/5 type UA.
    # Including the MicrosoftAjax pair lets us check errors quickly. In reality
    # we don't care about the output of this POST, just that it doesn't fail.
    report = client.post(
        run_report,
        data=form_data,
        headers={"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)", "X-MicrosoftAjax": "Delta=true"},
    )

    # Check error state
    _error_status(report, msg="Updating report locations failed!")
    if "compass.scouts.org.uk%2fError.aspx|" in report.text:
        raise ci.CompassReportError("Compass Error!")


def _form_data_appointments(form_data: dict[str, str], tree: html.HtmlElement) -> dict[str, str | None]:
    additional_form_data = {
        "ReportViewer1$ctl10": "ltr",
        "ReportViewer1$ctl11": "standards",
        "ReportViewer1$ctl05$ctl00$CurrentPage": "1",
        "ReportViewer1$ctl09$VisibilityState$ctl00": "ReportPage",
        "__EVENTTARGET": "ReportViewer1$ctl04$ctl07",
        "__EVENTARGUMENT": None,
        "__LASTFOCUS": None,
        "__ASYNCPOST": "true",
    }  # TODO this may not be needed. Test.

    # ReportViewer1$ctl04$ctl05$txtValue - County Label
    # ReportViewer1$ctl04$ctl07$txtValue - District Label
    # ReportViewer1$ctl04$ctl09$txtValue - Role Statuses
    # ReportViewer1$ctl04$ctl15$txtValue - Columns Label

    numbered_counties = _parse_drop_down_list(tree, "ReportViewer1_ctl04_ctl05_divDropDown")  # Counties
    numbered_districts = _parse_drop_down_list(tree, "ReportViewer1_ctl04_ctl07_divDropDown")  # Districts
    numbered_role_statuses = _parse_drop_down_list(tree, "ReportViewer1_ctl04_ctl09_divDropDown")  # Role Statuses
    numbered_column_names = _parse_drop_down_list(tree, "ReportViewer1_ctl04_ctl15_divDropDown")  # Report Fields

    # # Export regional roles only
    # form_data["ReportViewer1$ctl04$ctl05$txtValue"] = "Regional Roles"
    # form_data["ReportViewer1$ctl04$ctl05$divDropDown$ctl01$HiddenIndices"] = "0"

    # Export all districts
    form_data["ReportViewer1$ctl04$ctl05$txtValue"] = ", ".join(numbered_counties.values())
    form_data["ReportViewer1$ctl04$ctl05$divDropDown$ctl01$HiddenIndices"] = ",".join(numbered_counties.keys())
    form_data["ReportViewer1$ctl04$ctl07$txtValue"] = ", ".join(numbered_districts.values())
    form_data["ReportViewer1$ctl04$ctl07$divDropDown$ctl01$HiddenIndices"] = ",".join(numbered_districts.keys())

    # TODO this may not be needed. Test.
    # update text values of role statuses and column names from default indices
    form_data["ReportViewer1$ctl04$ctl09$txtValue"] = _get_defaults_labels(
        form_data, "ReportViewer1$ctl04$ctl09$divDropDown$ctl01$HiddenIndices", numbered_role_statuses
    )
    form_data["ReportViewer1$ctl04$ctl15$txtValue"] = _get_defaults_labels(
        form_data, "ReportViewer1$ctl04$ctl15$divDropDown$ctl01$HiddenIndices", numbered_column_names
    )

    return form_data | additional_form_data


def _extract_report_export_url(report_page: str) -> str:
    start = report_page.index("ExportUrlBase")
    cut = report_page[start:].removeprefix('ExportUrlBase":"')
    end = cut.index('"')
    full_url = cut[:end].encode().decode("unicode-escape")
    return f"{full_url}CSV"


def _download_report(client: Client, url_path: str, streaming: bool, filename: str | None = None) -> bytes:
    start = time.time()
    url = f"{Settings.base_url}/{url_path}"

    # actually do the download
    if streaming:
        csv_export = b""
        with client.get(url, stream=True) as r:
            _error_status(r)
            for chunk in r.iter_content(chunk_size=None):  # Chunk size == 1MiB
                csv_export += chunk
    else:
        csv_export = client.get(url).content

    logger.debug(f"Exporting took {time.time() - start:.2f}s")

    # maybe save to disk
    if filename is not None:
        logger.info("Saving report")
        with context_managers.filesystem_guard("Unable to write report export"):
            Path(filename).write_bytes(csv_export)
        logger.info("Report Saved")

    return csv_export


def _error_status(response: requests.Response, /, msg: str = "Request to Compass failed!") -> None:
    try:
        response.raise_for_status()
    except requests.HTTPError as err:
        raise ci.CompassNetworkError(msg) from err


def _parse_drop_down_list(tree: html.HtmlElement, element_id: str, /) -> dict[str, str]:
    table = tree.get_element_by_id(element_id)[0][0]
    return {str(i): row[0][0][0][1].text.replace("\xa0", " ") for i, row in enumerate(table[1:])}


def _get_defaults_labels(form_data: dict[str, str], default_indices_key: str, labels_map: dict[str, str]) -> str:
    indices = form_data.get(default_indices_key, "").split(",")
    return ", ".join(labels_map[i] for i in indices)


def _report_keep_alive(client: Client, report_page: str) -> str:
    logger.info(f"Extending Report Session {time.strftime('%Y-%m-%d %H-%M-%S')}")
    keep_alive_encoded = re.search(r'"KeepAliveUrl":"(.*?)"', report_page).group(1)  # type: ignore[union-attr]
    keep_alive = keep_alive_encoded.encode().decode("unicode-escape")
    response = client.post(f"{Settings.base_url}{keep_alive}")  # NoQA: F841 (unused variable)

    return keep_alive  # response



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

import pydantic

from compass.core.logger import logger
from compass.core.settings import Settings

if TYPE_CHECKING:
    from collections.abc import Iterator


@contextlib.contextmanager
def filesystem_guard(msg: str) -> Iterator[None]:
    try:
        yield
    except IOError as err:
        logger.error(f"{msg}: {err.errno} - {err.strerror}")


@contextlib.contextmanager
def validation_errors_logging(id_value: int, name: str = "Member No") -> Iterator[None]:
    try:
        yield
    except pydantic.ValidationError as err:
        if Settings.validation_errors is True:
            logger.error(f"Parsing Error! {name}: {id_value}")  # don't duplicate tracebacks
            raise err
        logger.exception(f"Parsing Error! {name}: {id_value}")



from __future__ import annotations

import datetime
from typing import Literal, Optional, TypedDict, Union
import warnings

import phonenumbers
import pydantic

# sync with _scrapers/member_profile.py::ROLE_STATUSES
TYPES_ROLE_STATUS = Literal["Cancelled", "Closed", "Full", "Pre provisional", "Provisional"]
TYPES_SEX = Literal["Male", "Female", "Unknown"]
TYPES_ETHNICITY = Literal[
    "1.English/Welsh/Scottish/Northern Irish/British",
    "2.Irish",
    "3.Gypsy or Irish Traveller",
    "4.Any other White background",
    "5.White and Black Caribbean",
    "6.White and Black African",
    "7.White and Asian",
    "8.Any other mixed or Multiple ethic group",
    "9.Indian",
    "10.Pakistani",
    "11.Bangladeshi",
    "12.Chinese",
    "13.Any other Asian Background",
    "14.African",
    "15.Caribbean",
    "16.Any other Black/African/Caribbean background",
    "17.Arab",
    "18.Other",
    "19.Prefer not to say",
]
TYPES_RELIGION = Literal[
    "Buddhist",
    "Christian (including all Christian denominations)",
    "Hindu",
    "Jewish",
    "Muslim",
    "Sikh",
    "Any other religion (please specify)",
    "No religion",
    "Prefer not to say",
]
TYPES_OCCUPATION = Literal[
    "Employed",
    "Unemployed",
    "Retired (whether receiving a pension or not)",
    "Student",
    "Long term sick or disabled",
    "Looking after home of family",
    "Other",
]
TYPES_ROLE_CLASS = Literal[
    "Administrator",
    "Advisor",
    "Assessor",
    "Co-ordinator",
    "Commissioner",
    "Committee",
    "Helper",
    "Honorary",
    "Leader",
    "Manager",
    "Secretary",
    "Staff",
    "System Role",
    "Supporter",
    "Trainer",
    "Default role class",
]
TYPES_REFERENCES = Literal[
    "Not Complete",
    "Not Required",
    "References Requested",
    "References Satisfactory",
    "References Unsatisfactory",
]
TYPES_DISCLOSURES_APPOINTMENT = Literal[
    "Application submitted - in progress",
    "Disclosure Issued",
    "ID check required",
    "No Disclosure",
]  # Disclosure statuses in role details popup
TYPES_LEARNING_METHOD = Literal[
    "Course",
    "DVD/Video",
    "E-Learning",
    "External Course",
    "Independent Study",
    "On the Job",
    "One to One",
    "Practical",
    "Small Group",
    "Workbook",
    "Your prior learning/experience recognised",
    "Other",
    # First Aid
    "Life Support",
    "Major illness",
    "Trauma and injury",
    # Manager and Supporter: Induction (rare)
    "RST Induction [England only]",
    # Manager and Supporter: Skills Courses
    "Skills Course - Achieving Growth",
    "Skills Course - Meeting the Challenges",
    "Skills Course - Skills of Management",
    # Manager and Supporter: Independent Learning
    "IL - Building Effective Teams",
    "IL - Dealing with Difficult Situations",
    "IL - Decision Making",
    "IL - Enabling Change",
    "IL - Financial and Physical Resources",
    "IL - Finding, Appointing and Welcoming Volunteers",
    "IL - Getting the Word Out",
    "IL - Keeping, Developing and Managing Volunteers",
    "IL - Leading Local Scouting",
    "IL - Managing Time and Personal Skills",
    "IL - Planning For Growth",
    "IL - Project Management",
    "IL - Safety for Managers and Supporters",
    "IL - Supporting the Adult Training Scheme",
]
TYPES_PERMIT = Literal[
    "Archery",
    "Bell Boating",
    "Canoeing",
    "Caving",
    "Climbing and Abseiling",
    "Dinghy Sailing",
    "Dragon Boating",
    "Hill Walking",
    "Hovercrafting",
    "Ice Climbing",
    "Kayaking",
    "Keelboating",
    "Kite Surfing",
    "Mine Exploration",
    "Motor Cruising",
    "Mountain Biking",
    "Narrow Boating",
    "Nights Away",
    "Personal Watercraft (Jet Ski)",
    "Power Boating",
    "Pulling",
    "Rafting (Traditional)",
    "Rowing and Sculling",
    "Scuba Diving",
    "Snorkelling",
    "Snowsports",
    "Water Skiing",
    "White Water Rafting",
    "Windsurfing",
    "Yachting",
]
TYPES_PERMIT_CATEGORIES = Literal[
    "B1 Waters",
    "B2 Waters",
    "B2+ Waters",
    "B3 Waters",
    "C Waters",
    "Campsite",
    "Greenfield",
    "Indoor",
    "Lightweight Expedition",
    "Open Inland B1 Waters",
    "Open Inland B2 Waters",
    "Open Inland B3 Waters",
    "River B1 Waters",
    "Sea B1 Waters",
    "Sea B2 Waters",
    "Surf B2 Waters",
    "With Compound Bows",
    "Without Compound Bows",
]
TYPES_PERMIT_TYPE = Literal["Leadership", "Supervisory"]
TYPES_AWARD_TYPE = Literal[
    "Chief Scout's 5 years Service Award",
    "Chief Scout's 10 years Service Award",
    "Chief Scout's 15 years Service Award",
    "Chief Scout's 20 years Service Award",
    "Chief Scout's 25 years Service Award",
    "Chief Scout's 30 years Service Award",
    "Chief Scout's 40 years Service Award",
    "Chief Scout's 50 years Service Award",
    "Chief Scout's 60 years Service Award",
    "Chief Scout's 70 years Service Award",
    # Local / less formal awards
    "Commissioners Commendation",
    # Formal awards process - lower good service
    "Chief Scout's Commendation for Good Service",
    "Award for Merit",
    "Bar to the Award for Merit",
    "Silver Acorn",
    # Formal awards process - higher good service
    "Bar to the Silver Acorn",
    "Silver Wolf",
    # Meritorious Conduct Awards
    "Chief Scout's Commendation for Meritorious Conduct",
    "The Medal for Meritorious Conduct",
    # Gallantry Awards
    "The Gilt Cross",
    "The Silver Cross",
    "The Bronze Cross",
    # Other Awards
    "The Cornwell Scout Badge",
    "The Chief Scout's Personal Award",
    # Past Awards (not in the current awards scheme)
    "Medal of Merit",
    "Bar to the Medal of Merit",
]
TYPES_DISCLOSURE_PROVIDERS = Literal[
    "Access NI",
    "Atlantic Data",
    "DBS Paper Application",
    "Disclosure Scotland",
    "Local Check",  # BSO
    "Other",
]
TYPES_DISCLOSURE_STATUSES = Literal[
    "Application Withdrawn",
    "Application received at HQ",
    "Application submitted - in progress",
    "Disclosure Expired",
    "Disclosure Issued",
    "Expired",
    "Final applicant information required",  # BSO
    "ID check required",
    "ID selection required",
]  # Disclosure statuses in disclosures tab


# TODO find a better place to put this - not a pydantic type...
class AddressData(TypedDict):
    unparsed_address: Optional[str]
    country: Optional[str]
    postcode: Optional[str]
    county: Optional[str]
    town: Optional[str]
    street: Optional[str]


class MemberBase(pydantic.BaseModel):
    membership_number: int


class MemberRoleBase(pydantic.BaseModel):
    # Role details
    role_number: int
    role_title: str  # role_name
    role_start: datetime.date
    role_status: TYPES_ROLE_STATUS
    review_date: Optional[datetime.date] = None


class MemberDetailsAddress(pydantic.BaseModel):
    unparsed_address: Optional[str] = None  # this is the unmodified address string
    country: Optional[str] = None
    postcode: Optional[str] = None
    county: Optional[str] = None
    town: Optional[str] = None
    street: Optional[str] = None


# Personal Details Tab
class MemberDetails(MemberBase):
    # Name
    name: str
    known_as: str
    forenames: Optional[str] = None
    surname: Optional[str] = None

    # Core personal information
    birth_date: Optional[datetime.date] = None
    sex: Optional[TYPES_SEX] = None
    nationality: Optional[str] = None  # literal? Big list...!
    ethnicity: Optional[TYPES_ETHNICITY] = None
    religion: Optional[TYPES_RELIGION] = None
    religion_detail: Optional[str] = None
    occupation: Optional[TYPES_OCCUPATION] = None
    occupation_detail: Optional[str] = None
    join_date: Optional[datetime.date] = None

    # Contact Details
    main_phone: Optional[str] = None
    main_email: Optional[pydantic.EmailStr] = None
    address: MemberDetailsAddress = MemberDetailsAddress()

    # Additional / miscellaneous details (maybe add keyed models ref #86)
    disabilities: Optional[dict[str, str]] = None
    qualifications: Optional[dict[str, str]] = None
    hobbies: Optional[dict[str, str]] = None

    # pylint doesn't support custom decorators. https://github.com/PyCQA/pylint/issues/1694
    # pylint: disable=no-self-use
    @pydantic.validator("main_phone")
    def check_phone_number(cls, v: Optional[str], values: dict[str, object]) -> Optional[str]:
        if v is None or not v or v == "0" or len(v) < 2:
            return None  # checks for v in (None, "", "0", .{1,2})

        try:
            n = phonenumbers.parse(v, "GB")
            if phonenumbers.is_valid_number(n):
                if n.country_code == 44:
                    return f"{phonenumbers.format_number(n, phonenumbers.PhoneNumberFormat.NATIONAL)}"
                return f"{phonenumbers.format_number(n, phonenumbers.PhoneNumberFormat.INTERNATIONAL)}"
        except phonenumbers.NumberParseException:
            pass

        membership_number = values["membership_number"]
        warnings.warn(f"Member No {membership_number}: phone number {v} is not valid!", RuntimeWarning)
        return None  # mypy being silly


# Roles Tab (Main List - item)
class MemberRoleCore(MemberBase, MemberRoleBase):
    # Role details
    role_end: Optional[datetime.date] = None
    role_class: TYPES_ROLE_CLASS
    role_type: Optional[str] = None
    # Full list of role types runs to 458, probably too many for a Literal type
    # xref: https://compass.scouts.org.uk/Roles.aspx?closed=Y

    # Location details
    location_id: Optional[int] = None
    location_name: Optional[str] = None

    # Is the role details popup accessible?
    can_view_details: bool


# Roles Tab (Main List - collection)
class MemberRolesCollection(pydantic.BaseModel):
    roles: dict[int, MemberRoleCore]
    membership_duration: float  # Membership duration in qualifying roles, in years
    primary_role: Optional[int]  # Primary role number. None if no primary role found


# Roles Tab (Role Detail Popup - Main)
class MemberRoleDetail(MemberBase, MemberRoleBase):
    organisation_level: str
    birth_date: Optional[datetime.date]
    name: str

    # Approval Process
    line_manager_number: Optional[int] = None
    line_manager: Optional[str] = None

    # Approval information
    ce_check: Optional[datetime.date]  # Optional for Closed roles - e.g. #499, role closed 1976, or if Pending
    disclosure_check: Optional[TYPES_DISCLOSURES_APPOINTMENT]
    disclosure_date: Optional[datetime.date]
    references: Optional[TYPES_REFERENCES] = None
    appointment_panel_approval: Optional[Literal["NC", "NR", "S", "U"]] = None
    commissioner_approval: Optional[Literal["NC", "NR", "RR", "S", "U"]] = None
    committee_approval: Optional[Literal["NC", "S", "U"]] = None


# Roles Tab (Role Detail Popup - Hierarchy)
class MemberRoleHierarchy(pydantic.BaseModel):
    # Role location
    organisation: str
    country: Optional[str] = None
    region: Optional[str] = None
    county: Optional[str] = None
    district: Optional[str] = None
    group: Optional[str] = None
    section: Optional[str] = None


# Roles Tab (Role Detail Popup - Getting Started)
class MemberRoleGettingStartedModule(pydantic.BaseModel):
    validated: Optional[datetime.date] = None
    validated_by: Optional[str] = None


# Roles Tab (Role Detail Popup - Getting Started)
class MemberRoleGettingStarted(pydantic.BaseModel):
    """Getting Started Training."""

    module_01: Optional[MemberRoleGettingStartedModule] = None
    trustee_intro: Optional[MemberRoleGettingStartedModule] = None
    module_02: Optional[MemberRoleGettingStartedModule] = None
    module_03: Optional[MemberRoleGettingStartedModule] = None
    module_04: Optional[MemberRoleGettingStartedModule] = None
    gdpr: Optional[MemberRoleGettingStartedModule] = None
    safety: Optional[MemberRoleGettingStartedModule]
    safeguarding: Optional[MemberRoleGettingStartedModule]


# Roles Tab (Role Detail Popup - All)
class MemberRolePopup(pydantic.BaseModel):
    hierarchy: MemberRoleHierarchy
    details: MemberRoleDetail
    getting_started: MemberRoleGettingStarted


# Training Tab (Role)
class MemberTrainingRole(MemberRoleBase):
    # Location details
    location: Optional[str] = None

    # Training Advisor details
    ta_data: Optional[str]
    ta_number: Optional[int]
    ta_name: Optional[str]

    # Completion details
    completion: Optional[str]
    completion_type: Optional[Literal["PLP", "Wood Badge"]]
    completion_date: Optional[datetime.date]
    wood_badge_number: Optional[int]


# Training Tab (PLP)
class MemberTrainingPLP(pydantic.BaseModel):
    pk: int
    module_id: int
    code: str  # We could do literals for module name/code but this would be ~150 values
    name: str

    # Learning details
    learning_required: Optional[bool]
    learning_method: Optional[TYPES_LEARNING_METHOD]
    learning_completed: Optional[datetime.date]
    learning_date: Optional[datetime.date]

    # Validation details
    validated_membership_number: Optional[int]
    validated_name: Optional[str]
    validated_date: Optional[datetime.date]


# Training Tab (OGL - item)
class MemberMOGL(pydantic.BaseModel):
    completed_date: Optional[datetime.date]
    renewal_date: Optional[datetime.date]  # GDPR has no renewal date...??


# Training Tab (OGL - all)
class MemberMandatoryTraining(pydantic.BaseModel):
    safety: MemberMOGL
    safeguarding: MemberMOGL
    first_aid: MemberMOGL
    gdpr: MemberMOGL


# Training Tab
class MemberTrainingTab(pydantic.BaseModel):
    roles: dict[int, MemberTrainingRole]
    plps: dict[int, list[MemberTrainingPLP]]
    mandatory: MemberMandatoryTraining


# Permits Tab
class MemberPermit(MemberBase):
    permit_type: TYPES_PERMIT
    category: str
    type: TYPES_PERMIT_TYPE
    restrictions: str
    expires: Optional[datetime.date]
    status: Literal["PERM_OK", "PERM_EXP", "PERM_REV"]


class MemberAward(MemberBase):
    type: TYPES_AWARD_TYPE
    location: Optional[str]
    date: datetime.date


class MemberDisclosure(MemberBase):
    country: Optional[Literal["England & Wales", "Northern Ireland", "Overseas", "Scotland", "The Scout Association"]]
    provider: TYPES_DISCLOSURE_PROVIDERS
    type: Literal["Enhanced with Barring"]
    # If Application Withdrawn, no disclosure number. If Scottish in the early 2000s, 7 digits ending with an R
    number: Union[int, None, pydantic.constr(regex=r"^\d{7}R$")]  # type: ignore[valid-type]  # NoQA: F722
    issuer: Optional[TYPES_DISCLOSURE_PROVIDERS]
    issue_date: Optional[datetime.date]  # If Application Withdrawn, maybe no issue date
    status: TYPES_DISCLOSURE_STATUSES
    expiry_date: Optional[datetime.date]  # If Application Withdrawn, no expiry date



from __future__ import annotations

import datetime
from typing import Any, Final, Literal, Optional, TYPE_CHECKING, Union
import urllib.parse

from lxml import html

import compass.core as ci
from compass.core.logger import logger
from compass.core.settings import Settings
from compass.core.util import auth_header
from compass.core.util.client import Client

if TYPE_CHECKING:
    from collections.abc import Iterator

    _TYPES_STO = Literal[None, "0", "5", "X"]
    _TYPES_ROLES_DICT = dict[int, ci.TYPES_ROLE]
    _TYPES_LEVEL_MAP = dict[ci.TYPES_ORG_LEVELS, ci.TYPES_HIERARCHY_LEVELS]

_level_map: _TYPES_LEVEL_MAP = {
    "ORG": "Organisation",
    "ORST": "Organisation Section",
    "CNTR": "Country",
    "CNST": "Country Section",
    "REG": "Region",
    "RGST": "Regional Section",
    "CNTY": "County",  # Also Area/Scot Reg/Branch
    "CTST": "County Section",  # Also Area/Scot Reg/Branch
    "DIST": "District",
    "DTST": "District Section",
    "SGRP": "Group",
    "SGST": "Group Section",
}


class Logon:  # pylint: disable=too-many-instance-attributes
    """Create connection to Compass and authenticate. Holds session state.

    Logon flow is:
    1. Create a persistent state object (Session) to hold headers, cookies etc.
      a. Get ASP.NET session cookie from Compass. (HTTP request #1)
    2. Post login data to Compass. (HTTP request #2)
    3. Get sample page from Compass. (HTTP request #3)
    4. Verify login was successful
      a. Create dict of compass internal variables, e.g. Master.User.CN (POST_CTRL).
      b. Create dict of user's role title to internal role number
      c. Update data for authorisation headers for requests to Compass.
    IF Role title specified:
    5. Post new role number to Compass (HTTP request #4)
    6. Get sample page from Compass. (HTTP request #5)
    7. Get currently active role number, and check this equals the requested role number.
      a. Create dict of compass internal variables, e.g. Master.User.CN (POST_CTRL).
      b. Create dict of user's role title to internal role number
      c. Update data for authorisation headers for requests to Compass.

    """

    def __init__(
        self,
        *,
        client: Client,
        compass_props: ci.CompassProps,
        current_role: ci.TYPES_ROLE,
    ):
        """Constructor for Logon.

        We treat all properties as immutable after we leave init. Role can
        theoretically change in server-side state, but this is not supported
        behaviour.

        """
        self._client: Final[Client] = client

        self.compass_props: Final[ci.CompassProps] = compass_props
        self.current_role: Final[ci.TYPES_ROLE] = current_role
        user_props = compass_props.master.user

        # Default hierarchy level
        unit_number = user_props.on  # Organisation Number
        unit_level = user_props.lvl  # Level
        if unit_number is None or unit_level is None:
            raise ci.CompassError("Unit Number and Level must be specified!")
        self.hierarchy: Final = ci.HierarchyLevel(unit_id=unit_number, level=_level_map[unit_level])

        # User / role IDs
        if user_props.cn is None or user_props.mrn is None or user_props.jk is None:
            raise ci.CompassError("User IDs must be specified!")
        self.membership_number: Final[int] = user_props.cn
        self.role_number: Final[int] = user_props.mrn
        self._jk: Final[str] = user_props.jk  # ???? Key?  # Join Key??? SHA2-512

        # Session IDs
        self._asp_net_id: Final[str] = client.cookies["ASP.NET_SessionId"]
        if compass_props.master.sys.session_id is None:
            raise ci.CompassError("ASP.NET ID must be specified!")
        self._session_id: Final[str] = compass_props.master.sys.session_id

        # TODO session timeout logic

    @classmethod
    def from_logon(
        cls: type[Logon],
        credentials: tuple[str, str],
        role_to_use: Optional[str] = None,
        role_location: Optional[str] = None,
    ) -> Logon:
        """Initialise a Logon object with login information.

        Args:
            credentials: username and password
            role_to_use: Compass role to use
            role_location: Role location, if multiple identical role titles

        Returns:
            Initialised Logon object

        Raises:
            CompassNetworkError:
                For errors while executing the HTTP call
            CompassError:
                If initial connection to Compass fails
            CompassAuthenticationError:
                If authentication with Compass fails

        """
        client = _create_session()

        # Log in and try to confirm success
        props, roles = _logon_remote(client, credentials)

        if props.master.user.mrn is None:
            raise ci.CompassError("Role Number must be an integer!")

        if role_to_use is not None:
            # Session contains updated auth headers from role change
            current_role, _roles_dict, compass_props = _change_role(client, role_to_use, role_location)
            return cls(client=client, compass_props=compass_props, current_role=current_role)

        return cls(client=client, compass_props=props, current_role=roles[props.master.user.mrn])

    @classmethod
    def from_session(
        cls: type[Logon], asp_net_id: str, user_props: dict[str, Union[str, int]], session_id: str, current_role: ci.TYPES_ROLE
    ) -> Logon:
        """Initialise a Logon object with stored data.

        This method is used to avoid logging in many times, by enabling reuse
        of an existing sever-side session in Compass. It is used by the main
        compass-interface web API.

        Args:
            asp_net_id: ASP.NET Session ID, from cookie
            user_props: Compass master.sys.user properties
            session_id: Compass session UID
            current_role: Role used by initialised session

        Returns:
            Initialised Logon object

        """
        client = Client()
        client.cookies.set("ASP.NET_SessionId", asp_net_id, domain=Settings.base_domain)  # type: ignore[no-untyped-call]

        logon = cls(
            client=client,
            compass_props=ci.CompassProps.parse_obj({"master": {"user": dict(user_props), "sys": {"session_id": session_id}}}),
            current_role=current_role,
        )

        _update_auth_headers(client, logon.membership_number, logon.role_number, session_id)

        return logon

    def __repr__(self) -> str:
        """String representation of the Logon class."""
        return f"{self.__class__} Compass ID: {self.membership_number} ({' - '.join(self.current_role)})"

    def _extend_session_timeout(self, sto: _TYPES_STO = "0") -> str:
        # Session time out. 4 values: None (normal), 0 (STO prompt) 5 (Extension, arbitrary constant) X (Hard limit)
        logger.debug(f"Extending session length {datetime.datetime.now()}")
        # TODO check STO.js etc for what happens when STO is None/undefined
        response = auth_header.auth_header_get(
            (self.membership_number, self.role_number, self._jk),
            self._client,
            f"{Settings.web_service_path}/STO_CHK",
            params={"pExtend": sto},
        )
        return response.content.decode("utf-8")


def _change_role(
    client: Client,
    new_role: str,
    location: Optional[str] = None,
    roles_dict: Optional[_TYPES_ROLES_DICT] = None,
) -> tuple[ci.TYPES_ROLE, _TYPES_ROLES_DICT, ci.CompassProps]:
    """Returns new Logon object with new role.

    If the user has multiple roles with the same role title, the first is used,
    unless the location parameter is set, where the location is exactly matched.
    """
    logger.info("Changing role")

    new_role = new_role.strip()

    # If we don't have the roles dict, generate it.
    if roles_dict is None:
        _props, roles_dict = _check_login(client)

    # Change role to the specified role number
    if location is not None:
        location = location.strip()
        member_role_number = next(num for num, name in roles_dict.items() if name == (new_role, location))
    else:
        member_role_number = next(num for num, name in roles_dict.items() if name[0] == new_role)

    response = client.post(f"{Settings.base_url}/API/ChangeRole", json={"MRN": member_role_number})  # b"false"
    Settings.total_requests += 1
    logger.debug(f"Compass ChangeRole call returned: {response.json()}")

    # Confirm Compass is reporting the changed role number, update auth headers
    compass_props, roles_dict = _check_login(client, check_role_number=member_role_number)
    current_role = roles_dict[member_role_number]

    logger.info(f"Role updated successfully! Role is now {current_role[0]} ({current_role[1]}).")

    return current_role, roles_dict, compass_props


def _create_session() -> Client:
    """Create a session and get ASP.Net Session ID cookie from the compass server."""
    client = Client()

    client.head(f"{Settings.base_url}/")  # use .head() as only headers needed to grab session cookie
    Settings.total_requests += 1

    if not client.cookies:
        raise ci.CompassError(
            "Could not create a session with Compass. Please check that this programme can "
            "access Compass (including firewalls etc.) and that Compass is currently online. "
        )

    return client


def _logon_remote(client: Client, auth: tuple[str, str]) -> tuple[ci.CompassProps, _TYPES_ROLES_DICT]:
    """Log in to Compass and confirm success."""
    # Referer is genuinely needed otherwise login doesn't work
    headers = {"Referer": f"{Settings.base_url}/login/User/Login"}

    username, password = auth
    credentials = {
        "EM": f"{username}",  # assume email?
        "PW": f"{password}",  # password
        "ON": f"{Settings.org_number}",  # organisation number
    }

    # log in
    logger.info("Logging in")
    client.post(f"{Settings.base_url}/Login.ashx", headers=headers, data=credentials)

    # verify log in was successful
    props, roles = _check_login(client)

    return props, roles


def _check_login(client: Client, check_role_number: Optional[int] = None) -> tuple[ci.CompassProps, _TYPES_ROLES_DICT]:
    """Confirms success and updates authorisation."""
    # Test 'get' for an exemplar page that needs authorisation.
    portal_url = f"{Settings.base_url}/MemberProfile.aspx?Page=ROLES&TAB"
    response = client.get(portal_url)

    # # Response body is login page for failure (~8Kb), but success is a 300 byte page.
    # if int(post_response.headers.get("content-length", 901)) > 900:
    #     raise CompassAuthenticationError("Login has failed")
    # Naive check for error, Compass redirects to an error page when something goes wrong
    # TODO what is the error page URL - what do we expect? From memory Error.aspx
    if response.url != portal_url:
        raise ci.CompassAuthenticationError("Login has failed")

    # Create lxml html.FormElement
    form = html.fromstring(response.content).forms[0]

    # Update session dicts with new role
    compass_props = _create_compass_props(form)  # Updates MRN property etc.
    roles_dict = dict(_roles_iterator(form))

    master_props = compass_props.master
    if master_props.user.cn is None or master_props.user.mrn is None or master_props.sys.session_id is None:
        raise ci.CompassError("Login verification failed! User and Session IDs not found!")
    membership_number: int = master_props.user.cn
    role_number: int = master_props.user.mrn
    session_id: str = master_props.sys.session_id

    # Set auth headers for new role
    _update_auth_headers(client, membership_number, role_number, session_id)

    # Update current role properties
    current_role_title, current_role_location = roles_dict[role_number]
    logger.debug(f"Using Role: {current_role_title} ({current_role_location})")

    # Verify role number against test value
    if check_role_number is not None:
        logger.debug("Confirming role has been changed")
        # Check that the role has been changed to the desired role. If not, raise exception.
        if check_role_number != role_number:
            raise ci.CompassAuthenticationError("Role failed to update in Compass")

    return compass_props, roles_dict


def _create_compass_props(form_tree: html.FormElement) -> ci.CompassProps:
    """Create Compass info dict from FormElement."""
    compass_props: dict[str, Any] = {}
    compass_vars = form_tree.fields["ctl00$_POST_CTRL"]
    for pair in compass_vars.split("~"):
        key, value = pair.split("#", 1)
        cd_tmp = compass_props
        levels = key.split(".")
        for level in levels[:-1]:
            cd_tmp = cd_tmp.setdefault(level, {})
        cd_tmp[levels[-1]] = value  # int or str

    if "Sys" in compass_props.get("Master", {}):
        props_sys = compass_props["Master"]["Sys"]
        if "WebPath" in props_sys:
            props_sys["WebPath"] = urllib.parse.unquote(props_sys["WebPath"])
        if "HardTime" in props_sys:
            props_sys["HardTime"] = datetime.time.fromisoformat(props_sys["HardTime"].replace(".", ":"))

    return ci.CompassProps(**compass_props)


def _roles_iterator(form_tree: html.FormElement) -> Iterator[tuple[int, ci.TYPES_ROLE]]:
    """Generate role number to role name mapping."""
    roles_rows = form_tree.xpath("//tbody/tr")  # get roles from compass page (list of table rows (tr))
    for row in roles_rows:
        if "Full" in row[5].text_content():  # from looking at the compass support group, only `Full` roles can be selected
            yield int(row.get("data-pk")), (row[0].text_content().strip(), row[2].text_content().strip())


def _update_auth_headers(
    client: Client,
    membership_number: int,
    role_number: int,
    session_id: str,
) -> None:
    client.headers.update(
        Authorization=f"{membership_number}~{role_number}",
        SID=session_id,  # Session ID
    )
